#
# Unit Tests for Elements of a Model
#
# TestSimpleVar                Class for testing single variables
# TestArrayVar                Class for testing array of variables
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from coopr.pyomo.base import IntegerSet
from coopr.pyomo import *
from coopr.opt import *
import pyutilib.th as unittest
import pyutilib.services

class PyomoModel(unittest.TestCase):

    def setUp(self):
        self.model = AbstractModel()
        self.instance = None

    def construct(self,filename=None):
        if filename is not None:
            self.instance = self.model.create(filename)
        else:
            self.instance = self.model.create()

class TestSimpleVar(PyomoModel):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)

    def tearDown(self):
        pass

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x.fixed, True)

    def test_fix_all(self):
        """Test fix all variables method"""
        self.model.A = RangeSet(4)
        self.model.x = Var()
        self.model.y = Var(self.model.A)

        self.instance = self.model.create()
        self.instance.fix_all_vars()
        self.assertEqual(self.instance.x.fixed, True)

        for a in self.instance.A:
            self.assertEqual(self.instance.y[a].fixed, True)

    def test_unfix_all(self):
        """Test unfix all variables method"""
        self.model.A = RangeSet(4)
        self.model.x = Var()
        self.model.y = Var(self.model.A)

        self.instance = self.model.create()
        self.instance.x.fixed = True
        for a in self.instance.A:
            self.instance.y[a].fixed = True
        self.instance.unfix_all_vars()

        self.assertEqual(self.instance.x.fixed, False)
        for a in self.instance.A:
            self.assertEqual(self.instance.y[a].fixed, False)


    def test_fix_indexed(self):
        """Test fix variables method"""
        self.model.A = RangeSet(4)
        self.model.y = Var(self.model.A)

        self.instance = self.model.create()
        self.instance.y.fix()

        for a in self.instance.A:
            self.assertEqual(self.instance.y[a].fixed, True)

    def test_unfix_indexed(self):
        """Test unfix variables method"""
        self.model.A = RangeSet(4)
        self.model.y = Var(self.model.A)

        self.instance = self.model.create()
        for a in self.instance.A:
            self.instance.y[a].fixed = True
        self.instance.unfix_all_vars()

        for a in self.instance.A:
            self.assertEqual(self.instance.y[a].fixed, False)


    def test_fix_nonindexed(self):
        """Test fix variables method"""
        self.model.A = RangeSet(4)
        self.model.x = Var()

        self.instance = self.model.create()
        self.instance.x.fix()
        self.assertEqual(self.instance.x.fixed, True)

    def test_unfix_nonindexed(self):
        """Test unfix variables method"""
        self.model.A = RangeSet(4)
        self.model.x = Var()
        self.model.y = Var(self.model.A)

        self.instance = self.model.create()
        self.instance.x.fixed = True
        self.instance.x.unfix()

        self.assertEqual(self.instance.x.fixed, False)


    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.instance.x.value = 3.5
        self.assertEqual(self.instance.x.value, 3.5)

    def test_initial_attr(self):
        """Test initial attribute"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.instance.x.initial = 3.5
        self.assertEqual(self.instance.x.initial, 3.5)

    def test_domain_attr(self):
        """Test domain attribute"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.instance.x.domain = Integers
        self.assertEqual(type(self.instance.x.domain), IntegerSet)
        self.assertEqual(self.instance.x.is_integer(), True)
        self.assertEqual(self.instance.x.is_binary(), False)
        self.assertEqual(self.instance.x.is_continuous(), False)

    def test_name_attr(self):
        """Test name attribute"""
        #
        # A user would never need to do this, but this
        # attribute is needed within Pyomo
        #
        self.model.x = Var()
        self.model.x.name = "foo"
        self.assertEqual(self.model.x.name, "foo")

    def test_lb_attr1(self):
        """Test lb attribute"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.instance.x.setlb(-1.0)
        self.assertEqual(value(self.instance.x.lb), -1.0)

    def test_lb_attr2(self):
        """Test lb attribute"""
        self.model.x = Var(within=NonNegativeReals, bounds=(-1,2))
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x.lb), 0.0)
        self.assertEqual(value(self.instance.x.ub), 2.0)

    def test_lb_attr3(self):
        """Test lb attribute"""
        self.model.p = Param(mutable=True, initialize=1)
        self.model.x = Var(within=NonNegativeReals, bounds=(self.model.p,None))
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x.lb), 1.0)
        self.instance.p = 2
        self.assertEqual(value(self.instance.x.lb), 2.0)

    def test_ub_attr1(self):
        """Test ub attribute"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.instance.x.setub(1.0)
        self.assertEqual(value(self.instance.x.ub), 1.0)

    def test_ub_attr2(self):
        """Test ub attribute"""
        self.model.x = Var(within=NonPositiveReals, bounds=(-2,1))
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x.lb), -2.0)
        self.assertEqual(value(self.instance.x.ub), 0.0)

    def test_within_option(self):
        """Test within option"""
        self.model.x = Var(within=Reals)
        self.construct()
        self.assertEqual(type(self.instance.x.domain), RealSet)
        self.assertEqual(self.instance.x.is_integer(), False)
        self.assertEqual(self.instance.x.is_binary(), False)
        self.assertEqual(self.instance.x.is_continuous(), True)

    def test_bounds_option1(self):
        """Test bounds option"""
        def x_bounds(model):
            return (-1.0,1.0)
        self.model.x = Var(bounds=x_bounds)
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x.lb), -1.0)
        self.assertEqual(value(self.instance.x.ub), 1.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = Var(bounds=(-1.0,1.0))
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x.lb), -1.0)
        self.assertEqual(value(self.instance.x.ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""
        def x_init(model):
            return 1.3
        self.model.x = Var(initialize=x_init)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x.value, 1.3)
        self.assertEqual(self.instance.x.initial, 1.3)

    def test_initialize_reset_with_function(self):
        """Test initialize option / reset method with an initialization rule"""
        def init_rule(model):
            return 1.3
        self.model.x = Var(initialize=init_rule)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x, 1.3)
        self.instance.x = 1
        self.assertEqual(self.instance.x, 1)
        self.instance.reset()
        self.assertEqual(self.instance.x, 1.3)

    def test_initialize_reset_with_dict(self):
        """Test initialize option / reset method with a dictionary"""
        self.model.x = Var(initialize={None:1.3})
        self.instance = self.model.create()
        self.assertEqual(self.instance.x, 1.3)
        self.instance.x = 1
        self.assertEqual(self.instance.x, 1)
        self.instance.reset()
        self.assertEqual(self.instance.x, 1.3)

    def test_initialize_reset_with_const(self):
        """Test initialize option / reset method with a constant"""
        self.model.x = Var(initialize=1.3)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x, 1.3)
        self.instance.x = 1
        self.assertEqual(self.instance.x, 1)
        self.instance.reset()
        self.assertEqual(self.instance.x, 1.3)

    def test_reset_without_initial_value(self):
        """Test reset method"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.assertEqual(self.instance.x.initial,None)
        self.assertEqual(self.instance.x.value,None)
        self.instance.x = 5
        self.assertEqual(self.instance.x.initial,None)
        self.assertEqual(self.instance.x.value,5)
        self.instance.x.reset()
        self.assertEqual(self.instance.x.initial,None)
        self.assertEqual(self.instance.x.value,None)

    def test_dim(self):
        """Test dim method"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.assertEqual(self.instance.x.dim(),0)

    def test_keys(self):
        """Test keys method"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.assertEqual(list(self.instance.x.keys()),[None])

    def test_len(self):
        """Test len method"""
        self.model.x = Var()
        self.instance = self.model.create()
        self.assertEqual(len(self.instance.x),1)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = Var(initialize=3.3)
        self.instance = self.model.create()
        tmp = value(self.instance.x.initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = float(self.instance.x.initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = int(self.instance.x.initial)
        self.assertEqual( type(tmp), int)
        self.assertEqual( tmp, 3 )


class TestArrayVar(TestSimpleVar):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        self.model.A = Set(initialize=[1,2])

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = Var(self.model.A)
        self.model.y = Var(self.model.A)
        self.instance = self.model.create()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x[1].fixed, False)
        self.instance.y[1].fixed=True
        self.assertEqual(self.instance.y[1].fixed, True)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = Var(self.model.A)
        self.model.y = Var(self.model.A)
        self.instance = self.model.create()
        try:
            self.instance.x = 3.5
            self.fail("Expected ValueError")
        except ValueError:
            pass
        self.instance.y[1] = 3.5
        self.assertEqual(self.instance.y[1], 3.5)

    #def test_initial_attr(self):
        #"""Test initial attribute"""
        #self.model.x = Var(self.model.A)
        #self.instance = self.model.create()
        #self.instance.x.initial = 3.5
        #self.assertEqual(self.instance.x[1].initial, 3.5)

    #def test_lb_attr(self):
        #"""Test lb attribute"""
        #self.model.x = Var(self.model.A)
        #self.instance = self.model.create()
        #self.instance.x.setlb(-1.0)
        #self.assertEqual(value(self.instance.x[1].lb), -1.0)

    #def test_ub_attr(self):
        #"""Test ub attribute"""
        #self.model.x = Var(self.model.A)
        #self.instance = self.model.create()
        #self.instance.x.setub(1.0)
        #self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_initialize_reset_with_function(self):
        """Test initialize option / reset method with an initialization rule"""
        def init_rule(model, key):
            i = key+11
            return key == 1 and 1.3 or 2.3
        self.model.x = Var(self.model.A,initialize=init_rule)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2], 2.3)

    def test_initialize_reset_with_dict(self):
        """Test initialize option / reset method with a dictionary"""
        self.model.x = Var(self.model.A,initialize={1:1.3,2:2.3})
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2], 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2], 2.3)

    def test_initialize_reset_with_subdict(self):
        """Test initialize option / reset method with a dictionary"""
        self.model.x = Var(self.model.A,initialize={1:1.3})
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2].value, None)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2].value, None)

    def test_initialize_reset_with_const(self):
        """Test initialize option / reset method with a constant"""
        self.model.x = Var(self.model.A,initialize=3)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1], 3)
        self.assertEqual(self.instance.x[2], 3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 3)
        self.assertEqual(self.instance.x[2], 3)

    def test_reset_without_initial_value(self):
        """Test reset method"""
        self.model.x = Var(self.model.A)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1].initial,None)
        self.assertEqual(self.instance.x[1].value,None)
        self.assertEqual(self.instance.x[2].initial,None)
        self.assertEqual(self.instance.x[2].value,None)
        self.instance.x[1] = 5
        self.instance.x[2] = 6
        self.assertEqual(self.instance.x[1].initial,None)
        self.assertEqual(self.instance.x[1].value,5)
        self.assertEqual(self.instance.x[2].initial,None)
        self.assertEqual(self.instance.x[2].value,6)
        self.instance.x.reset()
        self.assertEqual(self.instance.x[1].initial,None)
        self.assertEqual(self.instance.x[1].value,None)
        self.assertEqual(self.instance.x[2].initial,None)
        self.assertEqual(self.instance.x[2].value,None)

    def test_bounds_option1(self):
        """Test bounds option"""
        def x_bounds(model, i):
            return (-1.0,1.0)
        self.model.x = Var(self.model.A, bounds=x_bounds)
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = Var(self.model.A, bounds=(-1.0,1.0))
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""
        def x_init(model, i):
            return 1.3
        self.model.x = Var(self.model.A, initialize=x_init)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1].initial, 1.3)

    def test_dim(self):
        """Test dim method"""
        self.model.x = Var(self.model.A)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x.dim(),1)

    def test_keys(self):
        """Test keys method"""
        self.model.x = Var(self.model.A)
        self.instance = self.model.create()
        self.assertEqual(set(self.instance.x.keys()),set([1,2]))

    def test_len(self):
        """Test len method"""
        self.model.x = Var(self.model.A)
        self.instance = self.model.create()
        self.assertEqual(len(self.instance.x),2)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = Var(self.model.A,initialize=3.3)
        self.instance = self.model.create()
        tmp = value(self.instance.x[1].initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = float(self.instance.x[1].initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = int(self.instance.x[1].initial)
        self.assertEqual( type(tmp), int)
        self.assertEqual( tmp, 3 )


class TestVarList(PyomoModel):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = VarList()
        self.model.y = VarList()
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.y.add()
        self.instance.y.add()
        self.instance.y.add()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x[1].fixed, False)
        self.instance.y[1].fixed=True
        self.assertEqual(self.instance.y[1].fixed, True)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = VarList()
        self.model.y = VarList()
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.y.add()
        self.instance.y.add()
        self.instance.y.add()
        try:
            self.instance.x = 3.5
            self.fail("Expected ValueError")
        except ValueError:
            pass
        self.instance.y[1] = 3.5
        self.assertEqual(self.instance.y[1], 3.5)

    def test_initialize_reset_with_function(self):
        """Test initialize option / reset method with an initialization rule"""
        def init_rule(model, key):
            i = key+11
            return key == 1 and 1.3 or 2.3
        self.model.x = VarList(initialize=init_rule)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2], 2.3)

    def test_initialize_reset_with_dict(self):
        """Test initialize option / reset method with a dictionary"""
        self.model.x = VarList(initialize={1:1.3,2:2.3})
        self.model.x.add()
        self.model.x.add()
        self.model.x.add()
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2], 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2], 2.3)

    def test_initialize_reset_with_subdict(self):
        """Test initialize option / reset method with a dictionary"""
        self.model.x = VarList(initialize={1:1.3})
        self.model.x.add()
        self.model.x.add()
        self.model.x.add()
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2].value, None)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 1.3)
        self.assertEqual(self.instance.x[2].value, None)

    def test_initialize_reset_with_const(self):
        """Test initialize option / reset method with a constant"""
        self.model.x = VarList(initialize=3)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1], 3)
        self.assertEqual(self.instance.x[2], 3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1], 1)
        self.assertEqual(self.instance.x[2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1], 3)
        self.assertEqual(self.instance.x[2], 3)

    def test_reset_without_initial_value(self):
        """Test reset method"""
        self.model.x = VarList()
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].initial,None)
        self.assertEqual(self.instance.x[1].value,None)
        self.assertEqual(self.instance.x[2].initial,None)
        self.assertEqual(self.instance.x[2].value,None)
        self.instance.x[1] = 5
        self.instance.x[2] = 6
        self.assertEqual(self.instance.x[1].initial,None)
        self.assertEqual(self.instance.x[1].value,5)
        self.assertEqual(self.instance.x[2].initial,None)
        self.assertEqual(self.instance.x[2].value,6)
        self.instance.x.reset()
        self.assertEqual(self.instance.x[1].initial,None)
        self.assertEqual(self.instance.x[1].value,None)
        self.assertEqual(self.instance.x[2].initial,None)
        self.assertEqual(self.instance.x[2].value,None)

    def test_bounds_option1(self):
        """Test bounds option"""
        def x_bounds(model, i):
            return (-1.0,1.0)
        self.model.x = VarList(bounds=x_bounds)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = VarList(bounds=(-1.0,1.0))
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""
        def x_init(model, i):
            return 1.3
        self.model.x = VarList(initialize=x_init)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].initial, 1.3)

    def test_domain1(self):
        self.model.x = VarList(domain=NonNegativeReals)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(str(self.instance.x.domain), str(NonNegativeReals))
        self.assertEqual(str(self.instance.x[1].domain), str(NonNegativeReals))
        self.instance.x.domain = Integers
        self.assertEqual(str(self.instance.x.domain), str(Integers))

    def test_domain2(self):
        def x_domain(model, i):
            if i == 0:
                return NonNegativeReals
            elif i == 1:
                return Reals
            elif i == 2:
                return Integers
        self.model.x = VarList(domain=x_domain)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x.domain, None)
        self.assertEqual(str(self.instance.x[0].domain), str(NonNegativeReals))
        self.assertEqual(str(self.instance.x[1].domain), str(Reals))
        self.assertEqual(str(self.instance.x[2].domain), str(Integers))
        try:
            self.instance.x.domain = Reals
        except AttributeError:
            pass
        self.assertEqual(self.instance.x.domain, None)

    # VarList doesn't handle generators yet
    @unittest.expectedFailure
    def test_domain3(self):
        def x_domain(model):
            yield NonNegativeReals
            yield Reals
            yield Integers
        self.model.x = VarList(domain=x_domain)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x.domain, None)
        self.assertEqual(str(self.instance.x[0].domain), str(NonNegativeReals))
        self.assertEqual(str(self.instance.x[1].domain), str(Reals))
        self.assertEqual(str(self.instance.x[2].domain), str(Integers))
        try:
            self.instance.x.domain = Reals
        except AttributeError:
            pass
        self.assertEqual(self.instance.x.domain, None)

    def test_dim(self):
        """Test dim method"""
        self.model.x = VarList()
        self.instance = self.model.create()
        self.instance.x.add()
        self.assertEqual(self.instance.x.dim(),1)

    def test_keys(self):
        """Test keys method"""
        self.model.x = VarList()
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(set(self.instance.x.keys()),set([0,1]))

    def test_len(self):
        """Test len method"""
        self.model.x = VarList()
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(len(self.instance.x),2)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = VarList(initialize=3.3)
        self.instance = self.model.create()
        self.instance.x.add()
        self.instance.x.add()
        tmp = value(self.instance.x[1].initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = float(self.instance.x[1].initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = int(self.instance.x[1].initial)
        self.assertEqual( type(tmp), int)
        self.assertEqual( tmp, 3 )


class Test2DArrayVar(TestSimpleVar):

    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        self.model.A = Set(initialize=[1,2])

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = Var(self.model.A,self.model.A)
        self.model.y = Var(self.model.A,self.model.A)
        self.instance = self.model.create()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x[1,2].fixed, False)
        self.instance.y[1,2].fixed=True
        self.assertEqual(self.instance.y[1,2].fixed, True)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = Var(self.model.A,self.model.A)
        self.model.y = Var(self.model.A,self.model.A)
        self.instance = self.model.create()
        try:
            self.instance.x = 3.5
            self.fail("Expected ValueError")
        except ValueError:
            pass
        self.instance.y[1,2] = 3.5
        self.assertEqual(self.instance.y[1,2], 3.5)

    #def test_initial_attr(self):
        #"""Test initial attribute"""
        #self.model.x = Var(self.model.A,self.model.A)
        #self.instance = self.model.create()
        #self.instance.x.initial = 3.5
        #self.assertEqual(self.instance.x[1,1].initial, 3.5)

    #def test_lb_attr(self):
        #"""Test lb attribute"""
        #self.model.x = Var(self.model.A,self.model.A)
        #self.instance = self.model.create()
        #self.instance.x.setlb(-1.0)
        #self.assertEqual(value(self.instance.x[2,1].lb), -1.0)

    #def test_ub_attr(self):
        #"""Test ub attribute"""
        #self.model.x = Var(self.model.A,self.model.A)
        #self.instance = self.model.create()
        #self.instance.x.setub(1.0)
        #self.assertEqual(value(self.instance.x[2,1].ub), 1.0)

    def test_initialize_reset_with_function(self):
        """Test initialize option / reset method with an initialization rule"""
        def init_rule(model, key1, key2):
            i = key1+1
            return key1 == 1 and 1.3 or 2.3
        self.model.x = Var(self.model.A,self.model.A,initialize=init_rule)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1,1], 1.3)
        self.assertEqual(self.instance.x[2,2], 2.3)
        self.instance.x[1,1] = 1
        self.instance.x[2,2] = 2
        self.assertEqual(self.instance.x[1,1], 1)
        self.assertEqual(self.instance.x[2,2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1,1], 1.3)
        self.assertEqual(self.instance.x[2,2], 2.3)

    def test_initialize_reset_with_dict(self):
        """Test initialize option / reset method with a dictionary"""
        self.model.x = Var(self.model.A,self.model.A,
                           initialize={(1,1):1.3,(2,2):2.3})
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1,1], 1.3)
        self.assertEqual(self.instance.x[2,2], 2.3)
        self.instance.x[1,1] = 1
        self.instance.x[2,2] = 2
        self.assertEqual(self.instance.x[1,1], 1)
        self.assertEqual(self.instance.x[2,2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1,1], 1.3)
        self.assertEqual(self.instance.x[2,2], 2.3)

    def test_initialize_reset_with_const(self):
        """Test initialize option / reset method with a constant"""
        self.model.x = Var(self.model.A,self.model.A,initialize=3)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1,1], 3)
        self.assertEqual(self.instance.x[2,2], 3)
        self.instance.x[1,1] = 1
        self.instance.x[2,2] = 2
        self.assertEqual(self.instance.x[1,1], 1)
        self.assertEqual(self.instance.x[2,2], 2)
        self.instance.reset()
        self.assertEqual(self.instance.x[1,1], 3)
        self.assertEqual(self.instance.x[2,2], 3)

    def test_reset_without_initial_value(self):
        """Test reset method"""
        self.model.x = Var(self.model.A,self.model.A)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1,1].initial,None)
        self.assertEqual(self.instance.x[1,1].value,None)
        self.assertEqual(self.instance.x[2,2].initial,None)
        self.assertEqual(self.instance.x[2,2].value,None)
        self.instance.x[1,1] = 5
        self.instance.x[2,2] = 6
        self.assertEqual(self.instance.x[1,1].initial,None)
        self.assertEqual(self.instance.x[1,1].value,5)
        self.assertEqual(self.instance.x[2,2].initial,None)
        self.assertEqual(self.instance.x[2,2].value,6)
        self.instance.x.reset()
        self.assertEqual(self.instance.x[1,1].initial,None)
        self.assertEqual(self.instance.x[1,1].value,None)
        self.assertEqual(self.instance.x[2,2].initial,None)
        self.assertEqual(self.instance.x[2,2].value,None)

    def test_initialize_option(self):
        """Test initialize option"""
        self.model.x = Var(self.model.A,self.model.A,initialize={(1,1):1.3,(2,2):2.3})
        self.instance = self.model.create()
        self.instance.x.reset()
        self.assertEqual(self.instance.x[1,1], 1.3)
        self.assertEqual(self.instance.x[2,2], 2.3)
        try:
            value(self.instance.x[1,2])
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_bounds_option1(self):
        """Test bounds option"""
        def x_bounds(model, i, j):
            return (-1.0*(i+j),1.0*(i+j))
        self.model.x = Var(self.model.A, self.model.A, bounds=x_bounds)
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x[1,1].lb), -2.0)
        self.assertEqual(value(self.instance.x[1,2].ub), 3.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = Var(self.model.A, self.model.A, bounds=(-1.0,1.0))
        self.instance = self.model.create()
        self.assertEqual(value(self.instance.x[1,1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1,1].ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""
        def x_init(model, i, j):
            return 1.3
        self.model.x = Var(self.model.A, self.model.A, initialize=x_init)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x[1,2].initial, 1.3)

    def test_dim(self):
        """Test dim method"""
        self.model.x = Var(self.model.A,self.model.A)
        self.instance = self.model.create()
        self.assertEqual(self.instance.x.dim(),2)

    def test_keys(self):
        """Test keys method"""
        self.model.x = Var(self.model.A,self.model.A)
        self.instance = self.model.create()
        ans = [(1,1),(1,2),(2,1),(2,2)]
        self.assertEqual(list(sorted(self.instance.x.keys())),ans)

    def test_len(self):
        """Test len method"""
        self.model.x = Var(self.model.A,self.model.A)
        self.instance = self.model.create()
        self.assertEqual(len(self.instance.x),4)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = Var(self.model.A,self.model.A,initialize=3.3)
        self.instance = self.model.create()
        tmp = value(self.instance.x[1,1].initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = float(self.instance.x[1,1].initial)
        self.assertEqual( type(tmp), float)
        self.assertEqual( tmp, 3.3 )
        tmp = int(self.instance.x[1,1].initial)
        self.assertEqual( type(tmp), int)
        self.assertEqual( tmp, 3 )


class TestVarComplexArray(PyomoModel):

    def test_index1(self):
        self.model.A = Set(initialize=range(0,4))
        def B_index(model):
            for i in model.A:
                if i%2 == 0:
                    yield i
        def B_init(model, i, j):
            if j:
                return 2+i
            return -(2+i)
        self.model.B = Var(B_index, [True,False], initialize=B_init)
        self.instance = self.model.create()
        #self.instance.pprint()
        self.assertEqual(set(self.instance.B.keys()),set([(0,True),(2,True),(0,False),(2,False)]))
        self.instance.reset()
        self.assertEqual(self.instance.B[0,True],2)
        self.assertEqual(self.instance.B[0,False],-2)
        self.assertEqual(self.instance.B[2,True],4)
        self.assertEqual(self.instance.B[2,False],-4)

    def test_index2(self):
        self.model.A = Set(initialize=range(0,4))
        def B_index(model):
            for i in model.A:
                if i%2 == 0:
                    yield i-1, i
        B_index.dimen=2
        def B_init(model, k, i, j):
            if j:
                return (2+i)*k
            return -(2+i)*k
        self.model.B = Var(B_index, [True,False], initialize=B_init)
        self.instance = self.model.create()
        #self.instance.pprint()
        self.assertEqual(set(self.instance.B.keys()),set([(-1,0,True),(1,2,True),(-1,0,False),(1,2,False)]))
        self.instance.reset()
        self.assertEqual(self.instance.B[-1,0,True],-2)
        self.assertEqual(self.instance.B[-1,0,False],2)
        self.assertEqual(self.instance.B[1,2,True],4)
        self.assertEqual(self.instance.B[1,2,False],-4)


class MiscVarTests(pyutilib.th.TestCase):

    def test_error1(self):
        a = Var(name="a")
        try:
            a = Var(foo=1)
            self.fail("test_error1")
        except ValueError:
            pass

    def test_getattr1(self):
        """
        Verify the behavior of non-standard suffixes with simple variable
        """
        model = AbstractModel()
        model.a = Var()
        model.suffix = Suffix(datatype=Suffix.INT)
        instance = model.create()
        self.assertEqual(instance.suffix.getValue(instance.a),None)
        instance.suffix.setValue(instance.a,True)
        self.assertEqual(instance.suffix.getValue(instance.a),True)

    def test_getattr2(self):
        """
        Verify the behavior of non-standard suffixes with an array of variables
        """
        model = AbstractModel()
        model.X = Set(initialize=[1,3,5])
        model.a = Var(model.X)
        model.suffix = Suffix(datatype=Suffix.INT)
        try:
            self.assertEqual(model.a.suffix,None)
            self.fail("Expected AttributeError")
        except AttributeError:
            pass
        instance = model.create()
        self.assertEqual(instance.suffix.getValue(instance.a[1]),None)
        instance.suffix.setValue(instance.a[1], True)
        self.assertEqual(instance.suffix.getValue(instance.a[1]),True)

    def test_error2(self):
        try:
            model=AbstractModel()
            model.a = Var(initialize=[1,2,3])
            model.b = Var(model.a)
            self.fail("test_error2")
        except TypeError:
            pass

    def test_contains(self):
        model=AbstractModel()
        model.a = Set(initialize=[1,2,3])
        model.b = Var(model.a)
        instance = model.create()
        self.assertEqual(1 in instance.b,True)

    def test_float_int(self):
        model=AbstractModel()
        model.a = Set(initialize=[1,2,3])
        model.b = Var(model.a,initialize=1.1)
        model.c = Var(initialize=2.1)
        model.d = Var()
        instance = model.create()
        instance.reset()
        self.assertEqual(float(instance.b[1]),1.1)
        self.assertEqual(int(instance.b[1]),1)
        self.assertEqual(float(instance.c),2.1)
        self.assertEqual(int(instance.c),2)
        try:
            float(instance.d)
            self.fail("expected ValueError")
        except ValueError:
            pass
        try:
            int(instance.d)
            self.fail("expected ValueError")
        except ValueError:
            pass
        try:
            float(instance.b)
            self.fail("expected TypeError")
        except TypeError:
            pass
        try:
            int(instance.b)
            self.fail("expected TypeError")
        except TypeError:
            pass

    def test_set_get(self):
        model=AbstractModel()
        model.a = Set(initialize=[1,2,3])
        model.b = Var(model.a,initialize=1.1,within=PositiveReals)
        model.c = Var(initialize=2.1, within=PositiveReals)
        try:
            model.b = 2.2
            self.fail("can't set the value of an array variable")
        except ValueError:
            pass
        instance = model.create()
        try:
            instance.c[1]=2.2
            self.fail("can't use an index to set a scalar variable")
        except KeyError:
            pass
        try:
            instance.b[4]=2.2
            self.fail("can't set an array variable with a bad index")
        except KeyError:
            pass
        try:
            instance.b[3] = -2.2
            #print "HERE",type(instance)
            #print "HERE",type(instance.b[3])
            self.fail("can't set an array variable with a bad value")
        except ValueError:
            pass
        try:
            tmp = instance.c[3]
            self.fail("can't index a scalar variable")
        except KeyError:
            pass

        try:
            instance.c.set_value('a')
            self.fail("can't set a bad value for variable c")
        except ValueError:
            pass
        try:
            instance.c.set_value(-1.0)
            self.fail("can't set a bad value for variable c")
        except ValueError:
            pass

        try:
            instance.c.initial = 'a'
            instance.c.reset()
            self.fail("can't set a bad initial for variable c")
        except ValueError:
            pass
        try:
            instance.c.initial = -1.0
            instance.c.reset()
            self.fail("can't set a bad initial for variable c")
        except ValueError:
            pass

        #try:
            #instance.c.ub = 'a'
            #self.fail("can't set a bad ub for variable c")
        #except ValueError:
            #pass
        #try:
            #instance.c.ub = -1.0
            #self.fail("can't set a bad ub for variable c")
        #except ValueError:
            #pass

        #try:
            #instance.c.fixed = 'a'
            #self.fail("can't fix a variable with a non-boolean")
        #except ValueError:
            #pass


if __name__ == "__main__":
    unittest.main()



# A Suffix example for the collection of duals.
#
# This Pyomo example is formulated as a python script.
# To run this script execute the following command:
#
# $ coopr_python gurobi_ampl_iis.py
#
# Execution of this script requires that a solver with
# dual reporting capabilities (e.g., CPLEX, Gurobi, Cbc)
# is in the current search path for executables
# on this system. This example was tested using Gurobi
# Solver 5.1.0

import coopr.environ
from coopr.pyomo import *
from coopr.opt import SolverFactory

### Create the a solver plugin
solver = 'gurobi'
solver_io = 'lp' # Uses the LP file interface
stream_solver = False     # True prints solver output to screen
keepfiles =     False     # True prints intermediate file names (.nl,.sol,...) 
opt = SolverFactory(solver,solver_io=solver_io)

if opt is None:
    print("")
    print("ERROR: Unable to create solver plugin for %s "\
          "using the %s interface" % (solver, solver_io))
    print("")
    exit(1)

# import the simple example model containing a
# 'dual' IMPORT Suffix component
from duals_pyomo import model

### Generate the constraint expression trees if necessary
if solver_io != 'nl':
    # only required when not using the ASL (NL file) interface
    model.preprocess()
###

### Send the model to gurobi_ampl and collect the solution
# The solver plugin will scan the model for all active suffixes
# valid for importing, which it will store into the results object
results = opt.solve(model,
                    keepfiles=keepfiles,
                    tee=stream_solver)
model.load(results)

print("")
print("Dual Solution")
print("%s: %s" % (model.con, model.dual.getValue(model.con)))




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import os.path
import json
try:
    import yaml
    yaml_available = True
except ImportError:
    yaml_available = False

from pyutilib.misc import Options
from coopr.core.plugin import alias, Plugin, implements

from coopr.pyomo.base.plugin import IDataManager, DataManagerFactory


def detuplize(d):
    if type(d) in (list,set):
        ans = []
        for item in d:
            if type(item) in (list,tuple,set):
                ans.append(list(item))
            else:
                ans.append(item)
        return ans
    ans = {}
    for k,v in d.items():
        if type(k) is tuple:
            _d = ans
            j = 0
            for i in k:
                j += 1
                if j == len(k):
                    _d = _d.setdefault(i, v)
                else:
                    _d = _d.setdefault(i, {})
        else:
            ans[k] = v
    return ans

def tuplize(d, ret=None, ndx=None):
    if type(d) is list:
        ans = []
        for val in d:
            if type(val) is list:
                item = []
                for v in val:
                    try:
                        item.append(int(v))
                    except:
                        try:
                            item.append(float(v))
                        except:
                            item.append(v)
                item = tuple(item)
            else:
                try:
                    item = int(val)
                except:
                    try:
                        item = float(val)
                    except:
                        item = val
            ans.append(item)
        return ans
    elif not type(d) is dict:
        return d
    if ret is None:
        ret = {}
        ndx = []
    for k, v in d.items():
        try:
            k = int(k)
        except:
            try:
                k = float(k)
            except:
                pass
        ndx.append(k)
        if type(v) is dict:
            tuplize(v, ret, ndx)
        elif len(ndx) == 1:
            ret[ndx[0]] = v
        else:
            ret[tuple(ndx)] = v
        ndx.pop()
    return ret


class JSONDictionary(Plugin):

    alias("json", "Import data from a JSON file.")

    implements(IDataManager, service=False)

    def __init__(self):
        self._info = {}
        self.options = Options()

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")

    def close(self):
        pass

    def read(self):
        """
        This function loads data from a JSON file and tuplizes the nested
        dictionaries and lists of lists.
        """
        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)
        INPUT = open(self.filename, 'r')
        jdata = json.load(INPUT)
        self._info = {}
        for k,v in jdata.items():
            self._info[k] = tuplize(v)
        INPUT.close()
        return True

    def write(self, data):
        """
        This function creates a JSON file for the specified data.
        """
        OUTPUT = open(self.filename, 'w')
        jdata = {}
        if self.options.data is None:
            for k,v in data.items():
                jdata[k] = detuplize(v)
        elif type(self.options.data) in (list, tuple):
            for k in self.options.data:
                jdata[k] = detuplize(data[k])
        else:
            k = self.options.data
            jdata[k] = detuplize(data[k])
        json.dump(jdata, OUTPUT)
        OUTPUT.close()
        return True

    def process(self, model, data, default):
        """
        Set the data for the selected components
        """
        if not self.options.namespace in data:
            data[self.options.namespace] = {}
        #
        try:
            if self.options.data is None:
                for key in self._info:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            elif type(self.options.data) in (list, tuple):
                for key in self.options.data:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            else:
                key = self.options.data
                self._set_data(data, self.options.namespace, key, self._info[key])
        except KeyError:
            raise IOError("Data value for '%s' is not available in JSON file '%s'" % (key, self.filename))

    def _set_data(self, data, namespace, name, value):
        if type(value) is dict:
            data[namespace][name] = value
        else:
            data[namespace][name] = {None: value}

    def clear(self):
        self._info = {}



class YamlDictionary(Plugin):

    alias("yaml", "Import data from a YAML file.")

    implements(IDataManager, service=False)

    def __init__(self):
        self._info = {}
        self.options = Options()

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")

    def close(self):
        pass

    def read(self):
        """
        This function loads data from a JSON file and tuplizes the nested
        dictionaries and lists of lists.
        """
        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)
        INPUT = open(self.filename, 'r')
        jdata = yaml.load(INPUT)
        self._info = {}
        for k,v in jdata.items():
            self._info[k] = tuplize(v)
        INPUT.close()
        return True

    def write(self, data):
        """
        This function creates a JSON file for the specified data.
        """
        OUTPUT = open(self.filename, 'w')
        jdata = {}
        if self.options.data is None:
            for k,v in data.items():
                jdata[k] = detuplize(v)
        elif type(self.options.data) in (list, tuple):
            for k in self.options.data:
                jdata[k] = detuplize(data[k])
        else:
            k = self.options.data
            jdata[k] = detuplize(data[k])
        yaml.dump(jdata, OUTPUT)
        OUTPUT.close()
        return True

    def process(self, model, data, default):
        """
        Set the data for the selected components
        """
        if not self.options.namespace in data:
            data[self.options.namespace] = {}
        #
        try:
            if self.options.data is None:
                for key in self._info:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            elif type(self.options.data) in (list, tuple):
                for key in self.options.data:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            else:
                key = self.options.data
                self._set_data(data, self.options.namespace, key, self._info[key])
        except KeyError:
            raise IOError("Data value for '%s' is not available in JSON file '%s'" % (key, self.filename))

    def _set_data(self, data, namespace, name, value):
        if type(value) is dict:
            data[namespace][name] = value
        else:
            data[namespace][name] = {None: value}

    def clear(self):
        self._info = {}


if not yaml_available:
    DataManagerFactory.deactivate('yaml')




#
# Unit Tests for Piecewise
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from coopr.pyomo import *
import pyutilib.th as unittest
import pyutilib.services

class TestMiscPiecewise(unittest.TestCase):

    # test that activate and deactivate work
    def test_activate_deactivate_indexed(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        model.y = Var(model.s)
        model.x = Var(model.s,bounds=(-1,1))
        args = ([1],model.y,model.x)
        keywords = {'pw_pts':{1:[-1,0,1]},\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,i,x: x**2}
        model.c = Piecewise(*args,**keywords)
        self.assertTrue(len(model.c[1].subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c[1].active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, True)
        self.assertEqual(model.c[1].active, True)
        model.c[1].deactivate()
        self.assertTrue(len(model.c[1].subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c[1].active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, True)
        self.assertEqual(model.c[1].active, False)
        model.c[1].activate()
        self.assertTrue(len(model.c[1].subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c[1].active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, True)
        self.assertEqual(model.c[1].active, True)
        model.c.deactivate()
        self.assertTrue(len(model.c[1].subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c[1].active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, False)
        self.assertEqual(model.c[1].active, False)

    # test that activate and deactivate work
    def test_activate_deactivate_nonindexed(self):
        model = ConcreteModel()
        model.y = Var()
        model.x = Var(bounds=(-1,1))
        args = (model.y,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.c = Piecewise(*args,**keywords)
        self.assertTrue(len(model.c.subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c.active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, True)
        model.c.deactivate()
        self.assertTrue(len(model.c.subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c.active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, False)
        model.c.activate()
        self.assertTrue(len(model.c.subcomponents(Constraint)) > 0)
        self.assertTrue(len(model.c.active_subcomponents(Constraint)) > 0)
        self.assertEqual(model.c.active, True)

    # test that indexed Piecewise can handle
    # non-indexed vars
    def test_indexed_with_nonindexed_vars(self):
        model = ConcreteModel()
        model.range1 = Var()

        model.x = Var(bounds=(-1,1))
        args = ([1],model.range1,model.x)
        keywords = {'pw_pts':{1:[-1,0,1]},\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,i,x: x**2}
        model.con1 = Piecewise(*args,**keywords)

        model.range2 = Var([1])
        model.y = Var([1],bounds=(-1,1))
        args = ([1],model.range2,model.y)
        model.con2 = Piecewise(*args,**keywords)

        args = ([1],model.range2,model.y[1])
        model.con3 = Piecewise(*args,**keywords)

    # test that nonindexed Piecewise can handle
    # _VarData (e.g model.x[1]
    def test_nonindexed_with_indexed_vars(self):
        model = ConcreteModel()
        model.range = Var([1])

        model.x = Var([1],bounds=(-1,1))
        args = (model.range[1],model.x[1])
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con1 = Piecewise(*args,**keywords)

    # test that Piecewise can be initialized on
    # an abstract model
    def test_abstract_piecewise(self):
        model = AbstractModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        instance = model.create()

    # test that Piecewise can be initialized on
    # a concrete model
    def test_concrete_piecewise(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)


class TestInvalidPiecewise(unittest.TestCase):


    # test the that Piecewise component raises
    # an exception when the LOG/DLOG reps
    # are requested without a (2^n)+1 length
    # pw_pts list
    def test_dlog_bad_length(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'pw_repn':'DLOG',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)

        try:
            keywords['pw_pts'] = [-1,0,0.5,1]
            model.con3 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with DLOG an pw_pts list with length not "\
                      "equal to (2^n)+1.")

    # test the that Piecewise component raises
    # an exception when the LOG/DLOG reps
    # are requested without a (2^n)+1 length
    # pw_pts list
    def test_log_bad_length(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'pw_repn':'LOG',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)

        try:
            keywords['pw_pts'] = [-1,0,0.5,1]
            model.con3 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with LOG an pw_pts list with length not "\
                      "equal to (2^n)+1.")

    # test the that Piecewise component raises
    # an exception with an unsorted list of
    # domain points
    def test_unsorted_pw_pts(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)

        try:
            keywords['pw_pts'] = [0,-1,1]
            model.con3 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with unsorted domain points.")

    # test the that Piecewise component raises
    # an exception when initialized without
    # a valid f_rule keyword
    def test_bad_f_rules(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)

        try:
            del keywords['f_rule']
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "without a proper f_rule keyword.")

        try:
            keywords['f_rule'] = None
            model.con2 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "without a proper f_rule keyword.")

        try:
            keywords['f_rule'] = model.x
            model.con3 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "without a proper f_rule keyword.")

    # test the that Piecewise component raises
    # an exception if the domain variable arguments
    # are not actually Pyomo vars
    def test_bad_var_args(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        try:
            args = (None,model.x)
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "without Pyomo vars as variable args.")

        try:
            args = (model.range,None)
            model.con2 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "without Pyomo vars as variable args.")

    # test the that Piecewise component raises
    # an exception if the piecewise bound type is
    # a bad value
    def test_bad_bound_type(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        try:
            keywords['pw_constr_type'] = 1.0
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with invalid bound type.")
        try:
            del keywords['pw_constr_type']
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with invalid bound type.")

    # test the that Piecewise component raises
    # an exception if the piecewise representation
    # keyword has an invalid value
    def test_bad_repn(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        try:
            keywords['pw_repn'] = 1.0
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with invalid piecewise representation.")

    # test the that Piecewise component raises
    # an exception if the warning_tol
    # keyword has an invalid value
    def test_bad_warning_tol(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        try:
            keywords['warning_tol'] = None
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with invalid warning_tol.")

    # test the that Piecewise component raises
    # an exception if initialized with an invalid
    # number of args
    def test_bad_args_count(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        try:
            args = (model.range,)
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with less than two arguments.")

    # test the that Piecewise component raises
    # an exception if initialized with an unbounded
    # domain variable
    def test_unbounded_var(self):
        model = ConcreteModel()
        model.range = Var()

        model.x = Var(bounds=(-1,1))
        args = (model.range,model.x)
        keywords = {'pw_pts':[-1,0,1],\
                    'pw_constr_type':'EQ',\
                    'f_rule':lambda model,x: x**2}
        model.con = Piecewise(*args,**keywords)
        try:
            model.x.setlb(None)
            model.x.setub(None)
            model.con1 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with unbounded domain variable.")

        # Check that the unbounded_domain_var keyword works
        model.con1 = Piecewise(unbounded_domain_var=True,*args,**keywords)
            
        model.y = Var(bounds=(0,None))
        try:
            args = (model.range,model.y)
            model.con2 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with unbounded domain variable.")
        
        model.z = Var(bounds=(None,10))
        try:
            args = (model.range,model.z)
            model.con3 = Piecewise(*args,**keywords)
        except Exception:
            pass
        else:
            self.fail("Piecewise should fail when initialized "\
                      "with unbounded domain variable.")


if __name__ == "__main__":
    unittest.main()
     
        



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

# TODO
# . rename 'filter' to something else
# . confirm that filtering is efficient

__all__ = ['Set', 'set_options', 'simple_set_rule', 'SetOf']

from six import itervalues, iteritems
import logging
import sys
import types
import copy
import itertools
from weakref import ref as weakref_ref

import pyutilib.misc
from pyutilib.misc import flatten_tuple as pyutilib_misc_flatten_tuple

from coopr.pyomo.base.misc import apply_indexed_rule, apply_parameterized_indexed_rule
from coopr.pyomo.base.component import Component, register_component, ComponentData
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent, UnindexedComponent_set

try:                                    #pragma:nocover
    long
    intlist = [int, long, float]
except:                                 #pragma:nocover
    intlist = [int, float]
    xrange = range

logger = logging.getLogger('coopr.pyomo')


def process_setarg(arg):
    """
    Process argument and return an associated set object.

    This method is used by SparseIndexedComponent
    """
    if isinstance(arg,IndexedSet):
        # Argument is an indexed Set instance
        raise TypeError("Cannot index a component with an indexed set")
    elif isinstance(arg,Set):
        # Argument is a non-indexed Set instance
        return arg
    elif isinstance(arg,Component):
        # Argument is some other component
        raise TypeError("Cannot index a component with a non-set component")
    else:
        try:
            #
            # If the argument has a set_options attribute, then use 
            # it to initialize a set
            #
            options = getattr(arg,'set_options')
            options['initialize'] = arg
            return Set(**options)
        except:
            pass
    # Argument is assumed to be an initialization function
    return Set(initialize=arg)


def set_options(**kwds):
    """
    This is a decorator for set initializer functions.  This
    decorator allows an arbitrary dictionary of values to passed
    through to the set constructor.

    Examples:
        @set_options(dimen=3)
        def B_index(model):
            return [(i,i+1,i*i) for i in model.A]

        @set_options(domain=Integers)
        def B_index(model):
            return range(10)
    """
    def decorator(func):
        func.set_options = kwds
        return func
    return decorator


def simple_set_rule( fn ):
    """
    This is a decorator that translates None into Set.End.
    This supports a simpler syntax in set rules, though these can be
    more difficult to debug when errors occur.

    Example:

    @simple_set_rule
    def A_rule(model, i, j):
        ...
    """

    def wrapper_function ( *args, **kwargs ):
        value = fn( *args, **kwargs )
        if value is None:
            return Set.End
        return value
    return wrapper_function


class _SetData(ComponentData):
    """
    This class defines the data for an unordered set.

    Constructor Arguments:
        owner       The Set object that owns this data.
        bounds      A tuple of bounds for set values: (lower, upper)
    
    Public Class Attributes:
        value       The set values
        _bounds     The tuple of bound values
    """

    __slots__ = ('value', '_bounds')

    def __init__(self, owner, bounds):
        #
        # The following is equivalent to calling 
        # the base ComponentData constructor.
        #
        self._component = weakref_ref(owner)
        #
        self._clear()
        self._bounds = bounds

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        state = super(_SetData, self).__getstate__()
        for i in _SetData.__slots__:
            state[i] = getattr(self, i)
        return state

    # Note: because None of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    def __getitem__(self, key):
        """
        Return the specified member of the set.

        This method generates an exception because the set is unordered.
        """
        raise ValueError("Cannot index an unordered set '%s'" % self._component().cname())

    def bounds(self):
        """
        Return bounds information.  The default value is 'None', which
        indicates that this set does not contain bounds.  Otherwise, this is
        assumed to be a tuple: (lower, upper).
        """
        return self._bounds

    def data(self):
        """
        The underlying set data.

        Note that this method is preferred to the direct use of the
        'value' attribute in most cases.  The reason is that the 
        underlying set values may not be stored as a Python set() object.
        In fact, the underlying set values may not be explicitly stored
        in the Set() object at all!
        """
        return self.value

    def _clear(self):
        """
        Reset the set data
        """
        self.value = set()

    def _add(self, val, verify=True):
        """
        Add an element, and optionally verify that it is a valid type.

        The type verification is done by the owning component.
        """
        if verify:
            self._component()._verify(val)
        self.value.add(val)

    def _discard(self, val):
        """
        Discard an element of this set.  This does not return an error
        if the element does not already exist.
        """
        self.value.discard(val)

    def __len__(self):
        """
        Return the number of elements in the set.
        """
        return len(self.value)

    def __iter__(self):
        """
        Return an iterator for the set.
        """
        return self.value.__iter__()

    def __contains__(self, val):
        """
        Return True if the set contains a given value.
        """
        return val in self.value


class _OrderedSetData(ComponentData):
    """
    This class defines the data for an ordered set.

    Constructor Arguments:
        owner       The Set object that owns this data.
        bounds      A tuple of bounds for set values: (lower, upper)
    
    Public Class Attributes:
        value       The set values
        _bounds     The tuple of bound values
        order_dict  A dictionary that maps from element value to element id.
                        Indices in this dictionary start with 1 (not 0).

    The ordering supported in this class depends on the 'ordered' attribute
    of the owning component:
        InsertionOrder      The order_dict maps from the insertion order
                                back to the member of the value array.
        SortedOrder         The ordered attribute of the owning component can 
                                be used to define the sort order.  By default,
                                the Python ordering of the set types is used.
                                Note that a _stable_ sort method is required
                                if the discard method is used.
    """

    __slots__ = ('value', 'order_dict', '_bounds')

    def __init__(self, owner, bounds):
        #
        # The following is equivalent to calling 
        # the base ComponentData constructor.
        #
        self._component = weakref_ref(owner)
        #
        self._bounds = bounds
        self._clear()

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        state = super(_OrderedSetData, self).__getstate__()
        for i in _OrderedSetData.__slots__:
            state[i] = getattr(self, i)
        return state

    # Note: because None of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    def bounds(self):
        """
        Return bounds information.  The default value is 'None', which
        indicates that this set does not contain bounds.  Otherwise, this is
        assumed to be a tuple: (lower, upper).
        """
        return self._bounds

    def data(self):
        """
        Return the underlying set data.

        Note that this method returns a value that is different from the
        'value' attribute.  The underlying set values are not be stored 
        as a Python set() object.
        """
        return set(self.value)

    def _sort(self):
        """
        Sort the set using the 'ordered' attribute of the owning
        component.  This recreates the order_dict dictionary, which indicates
        that the set is sorted.
        """
        _sorter = self.component().ordered
        self.value.sort(None if _sorter is Set.SortedOrder else _sorter)
        self.order_dict = dict((j,i) for i,j in enumerate(self.value))

    def _clear(self):
        """
        Reset the set data
        """
        self.value = []
        if self.component().ordered is Set.InsertionOrder:
            self.order_dict = {}
        else:
            self.order_dict = None

    def _add(self, val, verify=True):
        """
        Add an element, and optionally verify that it is a valid type.

        The type verification is done by the owning component.
        """
        if verify:
            self._component()._verify(val)
        if self.component().ordered is Set.InsertionOrder:
            self.order_dict[val] = len(self.value)
        else:
            #
            # Denote that the set is unordered by setting
            # order_dict to None.
            #
            self.order_dict = None
        self.value.append(val)

    def _discard(self, val):
        """
        Discard an element of this set.  This does not return an error
        if the element does not already exist.
        """
        #
        # If the set has not been sorted yet, then a linear scan is the
        # most efficient way to find and remove the element.
        #
        if self.order_dict is None:
            try:
                self.value.remove(val)
            except ValueError:
                pass
        else:
            try:
                _id = self.order_dict.pop(val)
            except KeyError:
                return
            del self.value[_id]
            #
            # Update the order_dict: this assumes the user-specified sorter
            # (if one was used) is stable.
            #
            for i in xrange(_id,len(self.value)):
                self.order_dict[self.value[i]] = i

    def __len__(self):
        """
        Return the number of elements in the set.
        """
        return len(self.value)

    def __iter__(self):
        """
        Return an iterator for the set.
        """
        if self.order_dict is None:
            self._sort()
        return self.value.__iter__()

    def __contains__(self, val):
        """
        Return True if the set contains a given value.
        """
        if self.order_dict is None:
            # 
            # The set is unordered, so it is fastest to do a linear
            # scan of the set.  However, we assume that if the user 
            # is testing membership then they are likely to
            # do so again in the near future.  Hence, we sort the set
            # and test the index.
            #
            self._sort()
        return val in self.order_dict

    def first(self):
        """
        Return the first element of the set.
        """
        return self[1]

    def last(self):
        """
        Return the last element of the set.
        """
        return self[len(self)]

    def __getitem__(self, idx):
        """
        Return the specified member of the set.
        
        The public Set API is 1-based, even though the
        internal order_dict is (pythonically) 0-based.
        """
        if self.order_dict is None:
            self._sort()
        return self.value[idx-1]

    def member(self, idx):
        """
        Return the specified member of the set.
        """
        logger.warning("DEPRECATED: The ordered set method \"x.member[idx]\" "
                       "is deprecated and will be removed in Coopr 4.0.  "
                       "Use x[idx] instead.")
        return self[idx]

    def ord(self, match_element):
        """
        Return the position index of the input value.  The 
        position indices start at 1.
        """
        if self.order_dict is None:
            self._sort()
        try:
            return self.order_dict[match_element] + 1
        except IndexError:
            raise IndexError("Unknown input element="+str(match_element)+" provided as input to ord() method for set="+self.cname(True))

    def next(self, match_element, k=1):
        """
        Return the next element in the set.  The default behavior is to
        return the very next element.  The k option can specify how many
        steps are taken to get the next element.

	    If the next element is beyond the end of the set, then an
	    exception is raised.
        """
        try:
            element_position = self.ord(match_element)
        except IndexError:
            raise KeyError("Cannot obtain next() member of set="+self.cname(True)+"; input element="+str(match_element)+" is not a member of the set!")
        #
        try:
            return self[element_position+k]
        except KeyError:
            raise KeyError("Cannot obtain next() member of set="+self.cname(True)+"; failed to access item in position="+str(element_position+k))

    def nextw(self, match_element, k=1):
        """
        Return the next element in the set.  The default behavior is to
        return the very next element.  The k option can specify how many
        steps are taken to get the next element.

        If the next element goes beyond the end of the list of elements
        in the set, then this wraps around to the beginning of the list.
        """
        try:
            element_position = self.ord(match_element)
        except KeyError:
            raise KeyError("Cannot obtain nextw() member of set="+self.cname(True)+"; input element="+str(match_element)+" is not a member of the set!")
        #
        return self[(element_position+k-1) % len(self.value) + 1]

    def prev(self, match_element, k=1):
        """
        Return the previous element in the set.  The default behavior is to
        return the element immediately prior to the specified element.
        The k option can specify how many steps are taken to get the previous
        element.

	    If the previous element is before the start of the set,
	    then an exception is raised.
        """
        return self.next(match_element, k=-k)

    def prevw(self, match_element, k=1):
        """
        Return the previous element in the set.  The default behavior is to
        return the element immediately prior to the specified element.
        The k option can specify how many steps are taken to get the previous
        element.

        If the previous element is before the start of the set,
        then this wraps around to the end of the list.
        """
        return self.nextw(match_element, k=-k)


class _IndexedSetData(_SetData):
    """
    This class adds the __call__ method, which is expected for
    indexed component data.  But we omit this from _SetData because
    we do not want to treat scalar sets as functors.
    """

    def __call__(self):
        """
        Return the underlying set data.
        """
        return self.value

    def clear(self):
        """
        Reset this data.
        """
        self._clear()

    def add(self, val):
        """
        Add an element to the set.
        """
        self._add(val)

    def discard(self, val):
        """
        Discard an element from the set.
        """
        self._discard(val)


class _IndexedOrderedSetData(_OrderedSetData):
    """
    This class adds the __call__ method, which is expected for
    indexed component data.  But we omit this from _OrderedSetData because
    we do not want to treat scalar sets as functors.
    """

    def __call__(self):
        """
        Return the underlying set data.
        """
        return self.value

    def clear(self):
        """
        Reset this data.
        """
        self._clear()

    def add(self, val):
        """
        Add an element to the set.
        """
        self._add(val)

    def discard(self, val):
        """
        Discard an element from the set.
        """
        self._discard(val)


class Set(SparseIndexedComponent):
    """
    A set object that is used to index other Pyomo objects.

    This class has a similar look-and-feel as a Python set class.
    However, the set operations defined in this class return another
    abstract Set object.  This class contains a concrete set, which
    can be initialized by the load() method.

    Constructor Arguments:
        name            The name of the set
        doc             A text string describing this component
        within          A set that defines the type of values that can
                            be contained in this set
        domain          A set that defines the type of values that can
                            be contained in this set
        initialize      A dictionary or rule for setting up this set 
                            with existing model data
        validate        A rule for validating membership in this set.  This has 
                            the functional form:
                                f: data -> bool
                            and returns true if the data belongs in the set
        dimen           Specify the set's arity, or None if no arity is enforced
        virtual         If true, then this is a virtual set that does not
                            store data using the class dictionary
        bounds          A 2-tuple that specifies the range of possible set values.
        ordered         Specifies whether the set is ordered.  Possible values are:
                            False           Unordered
                            True            Ordered by insertion order
                            InsertionOrder  Ordered by insertion order
                            SortedOrder     Ordered by sort order
                            <function>      Ordered with this comparison function
        filter          A function that is used to filter set entries.

    Public class attributes:
        doc             A text string describing this component
        name            A name for this component
        initialize      A dictionary or rule for setting up this set 
                            with existing model data
        validate        A rule for validating membership in this set.
        ordered         Specifies whether the set is ordered.
        filter          A function that is used to filter set entries.
        domain          A set that defines the type of values that can
                            be contained in this set
        virtual         If True, then this set does not store data using the class dictionary
        concrete        If True, then this set contains elements. (TODO)
    """

    End             = (1003,)
    InsertionOrder  = (1004,)
    SortedOrder     = (1005,)

    def __new__(cls, *args, **kwds):
        if cls != Set:
            return super(Set, cls).__new__(cls)
        if args == ():
            if kwds.get('ordered',False) is False:
                return SimpleSet.__new__(SimpleSet)
            else:
                return OrderedSimpleSet.__new__(OrderedSimpleSet)
        else:
            return IndexedSet.__new__(IndexedSet)

    def __init__(self, *args, **kwds):
        #
        # Default keyword values
        #
        kwds.setdefault("name", "_unknown_")
        self.initialize = kwds.pop("initialize", None)
        self.validate   = kwds.pop("validate", None)
        self.ordered    = kwds.pop("ordered", False)
        self.filter     = kwds.pop("filter", None)
        self.domain     = kwds.pop("within", None)
        self.domain     = kwds.pop('domain', self.domain )
        #
        if self.ordered is True:
            self.ordered = Set.InsertionOrder

        # We can't access self.dimen after its been written, so we use
        # tmp_dimen until the end of __init__
        tmp_dimen = 0

        # Get dimen from domain, if possible
        if self.domain is not None:
            tmp_dimen = self.domain.dimen
        if self._bounds is None and not self.domain is None:
            self._bounds = copy.copy(self.domain._bounds)

        # Make sure dimen and implied dimensions don't conflict
        kwd_dimen = kwds.pop("dimen", 0)
        if kwd_dimen != 0:
            if self.domain is not None and tmp_dimen != kwd_dimen:
                raise ValueError(\
                      ("Value of keyword 'dimen', %s, differs from the " + \
                       "dimension of the superset '%s', %s") % \
                       (str(kwd_dimen), str(self.domain.name), str(tmp_dimen)))
            else:
                tmp_dimen = kwd_dimen

        kwds.setdefault('ctype', Set)
        SparseIndexedComponent.__init__(self, *args, **kwds)

        if tmp_dimen == 0:
            # We set the default to 1
            tmp_dimen = 1
        if self.initialize is not None:
            #
            # Convert generators to lists, since a generator is not copyable
            #
            if type(self.initialize) is types.GeneratorType:
                self.initialize = list(self.initialize)
            #
            # Try to guess dimen from the initialize list
            #
            if not tmp_dimen is None:
                tmp=0
                if type(self.initialize) is tuple:
                    tmp = len(self.initialize)
                elif type(self.initialize) is list and len(self.initialize) > 0 \
                         and type(self.initialize[0]) is tuple:
                    tmp = len(self.initialize[0])
                else:
                    tmp = getattr(self.initialize, 'dimen', tmp)
                if tmp != 0:
                    if kwd_dimen != 0 and tmp != kwd_dimen:
                        raise ValueError("Dimension argument differs from the data in the initialize list")
                    tmp_dimen = tmp

        self.dimen = tmp_dimen

    def reset(self):
        """
        Reset the set values.

        This action does not have an effect.
        """
        pass

    def _verify(self, element):
        """
        Verify that the element is valid for this set.
        """
        if self.domain is not None and element not in self.domain:
            raise ValueError(
                "The value=%s is not valid for set=%s\n"
                "because it is not within the domain=%s"
                % ( element, self.cname(True), self.domain.cname(True) ) )
        if self.validate is not None:
            flag = False
            try:
                if self._parent is not None:
                    flag = apply_indexed_rule(self, self.validate, self._parent(), element)
                else:
                    flag = apply_indexed_rule(self, self.validate, None, element)
            except:
                pass
            if not flag:
                raise ValueError("The value="+str(element)+" violates the validation rule of set="+self.cname(True))
        if not self.dimen is None:
            if self.dimen > 1 and type(element) is not tuple:
                
                raise ValueError("The value="+str(element)+" is not a tuple for set="+self.cname(True)+", which has dimen="+str(self.dimen))
            elif self.dimen == 1 and type(element) is tuple:
                raise ValueError("The value="+str(element)+" is a tuple for set="+self.cname(True)+", which has dimen="+str(self.dimen))
            elif type(element) is tuple and len(element) != self.dimen:
                raise ValueError("The value="+str(element)+" does not have dimension="+str(self.dimen)+", which is needed for set="+self.cname(True))
        return True


class SimpleSetBase(Set):
    """
    A derived Set object that contains a single set.
    """

    def __init__(self, *args, **kwds):
        self.virtual    = kwds.pop("virtual", False)
        self.concrete   = not self.virtual
        Set.__init__(self, *args, **kwds)

    def valid_model_component(self):
        """
        Return True if this can be used as a model component.
        """
        if self.virtual and not self.concrete:
            return False
        return True

    def clear(self):
        """
        Clear that data in this component.
        """
        if self.virtual:
            raise TypeError("Cannot clear virtual set object `"+self.cname(True)+"'")
        self._clear()

    def check_values(self):
        """
        Verify that the values in this set are valid.
        """
        if not self.concrete:
            return
        for val in self:
            self._verify(val)

    def add(self, *args):
        """
        Add one or more elements to a set.
        """
        if self.virtual:
            raise TypeError("Cannot add elements to virtual set `"+self.cname(True)+"'")
        for val in args:
            tmp = pyutilib_misc_flatten_tuple(val)
            self._verify(tmp)
            try:
                if tmp in self:
                    #
                    # Generate a warning, since we expect that users will not plan to
                    # re-add the same element to a set.
                    #
                    logger.warning("Element "+str(tmp)+" already exists in set "+self.cname(True)+"; no action taken.")
                    continue
                self._add(tmp, False)
            except TypeError:
                raise TypeError("Problem inserting "+str(tmp)+" into set "+self.cname(True))

    def remove(self, element):
        """
        Remove an element from the set.

        If the element is not a member, raise an error.
        """
        if self.virtual:
            raise KeyError("Cannot remove element `"+str(element)+"' from virtual set "+self.cname(True))
        if element not in self:
            raise KeyError("Cannot remove element `"+str(element)+"' from set "+self.cname(True))
        self._discard(element)

    def discard(self, element):
        """
        Remove an element from the set.

        If the element is not a member, do nothing.
        """
        if self.virtual:
            raise KeyError("Cannot discard element `"+str(element)+"' from virtual set "+self.cname(True))
        self._discard(element)

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        _ordered = self.ordered
        if type(_ordered) is bool:
            pass
        elif _ordered is Set.InsertionOrder:
            _ordered = 'Insertion'
        elif _ordered is Set.SortedOrder:
            _ordered = 'Sorted'
        else:
            _ordered = '{user}'
        return (
            [("Dim", self.dim()),
             ("Dimen", self.dimen),
             ("Size", len(self)),
             ("Domain", None if self.domain is None else self.domain.name),
             ("Ordered", _ordered),
             ("Bounds", self._bounds)],
            iteritems( {None: self} ), 
            None, 
            lambda k, v: [ 
                "Virtual" if not v.concrete or v.virtual \
                    else v.value if v.ordered \
                    else sorted(v) ] )

    def _set_repn(self, other):
        """
        Return a Set subset for 'other'
        """
        if isinstance(other, SimpleSet):
            return other
        if isinstance(other, OrderedSimpleSet):
            return other
        return SetOf(other)

    def __len__(self):
        """
        Return the number of elements in this set.
        """
        if not self.concrete:
            raise ValueError("The size of a non-concrete set is unknown")
        return len(self.value)

    def __iter__(self):
        """
        Return an iterator for the underlying set
        """
        if not self.concrete:
            raise TypeError("Cannot iterate over a non-concrete set '%s'" % self.cname(True))
        return self.value.__iter__()

    def __reversed__(self):
        """
        Return a reversed iterator
        """
        return reversed(self.__iter__())

    def __hash__(self):
        """
        Hash this object
        """
        return Set.__hash__(self)

    def __eq__(self,other):
        """
        Equality comparison
        """
        if id(self) == id(other):
            return True
        if other is None:
            return False
        try:
            tmp = self._set_repn(other)
        except:
            return False
        if self.virtual:
            if tmp.virtual:
                return hash(self) == hash(tmp)
            return False
        if tmp.virtual:
            return False
        if self.dimen != tmp.dimen:
            return False
        return self.data().__eq__( tmp.data() )

    def __ne__(self,other):
        """
        Inequality comparison
        """
        return not self.__eq__(other)

    def __contains__(self, element):
        """
        Return True if element is a member of this set.
        """
        #
        # If the element is a set, then see if this is a subset.
        # We first test if the element is a number or tuple, before
        # doing the expensive calls to isinstance().
        #
        element_t = type(element)
        if not element_t in intlist and element_t is not tuple and \
               (isinstance(element,SimpleSet) or isinstance(element,OrderedSimpleSet)):
            return element.issubset(self)
        # 
        # When dealing with a concrete set, just check if the element is
        # in the set. There is no need for extra validation.
        #        
        if self._constructed and self.concrete is True:
           return self._set_contains(element)
        #
        # If this is not a valid element, then return False
        #
        try:
            self._verify(element)
        except:
            return False
        #
        # If the validation rule is used then we do not actually
        # check whether the data is in self.value.
        #
        if self.validate is not None and not self.concrete:
            return True
        #
        # The final check: return true if self.concrete is False, since we should
        # have already validated this value.  The following, or at least one of
        # the execution paths - is probably redundant with the above.
        #
        return not self.concrete or self._set_contains(element)

    def _set_contains(self, element):
        """
        A wrapper function that tests if the element is in
        the data associated with a concrete set.
        """ 
        return element in self.value

    def isdisjoint(self, other):
        """
        Return True if the set has no elements in common with 'other'. 
        Sets are disjoint if and only if their intersection is the empty set.
        """
        other = self._set_repn(other)
        tmp = self & other
        for elt in tmp:
            return False
        return True
        
    def issubset(self,other):
        """
        Return True if the set is a subset of 'other'.
        """
        if not self.concrete:
            raise TypeError("ERROR: cannot perform \"issubset\" test because the current set is not a concrete set.")
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            raise ValueError("Cannot perform set operation with sets "+self.cname(True)+" and "+other.cname(True)+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen))
        for val in self:
            if val not in other:
                return False
        return True

    def issuperset(self, other):
        """
        Return True if the set is a superset of 'other'.

        Note that we do not simply call other.issubset(self) because
        'other' may not be a Set instance.
        """
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            raise ValueError("Cannot perform set operation with sets "+self.cname(True)+" and "+other.cname(True)+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen))
        if not other.concrete:
            raise TypeError("ERROR: cannot perform \"issuperset\" test because the target set is not a concrete set.")
        for val in other:
            if val not in self:
                return False
        return True

    def union(self, *args):
        """
        Return the union of this set with one or more sets.
        """
        tmp = self
        for arg in args:
            tmp = _SetUnion(tmp, arg)
        return tmp

    def intersection(self, *args):
        """
        Return the intersection of this set with one or more sets
        """
        tmp = self
        for arg in args:
            tmp = _SetIntersection(tmp, arg)
        return tmp

    def difference(self, *args):
        """
        Return the difference between this set with one or more sets
        """
        tmp = self
        for arg in args:
            tmp = _SetDifference(tmp, arg)
        return tmp

    def symmetric_difference(self, *args):
        """
        Return the symmetric difference of this set with one or more sets
        """
        tmp = self
        for arg in args:
            tmp = _SetSymmetricDifference(tmp, arg)
        return tmp

    def cross(self, *args):
        """
        Return the cross-product between this set and one or more sets
        """
        tmp = self
        for arg in args:
            tmp = _SetProduct(tmp, arg)
        return tmp

    # <= is equivalent to issubset
    # >= is equivalent to issuperset
    # |  is equivalent to union
    # &  is equivalent to intersection
    # -  is equivalent to difference
    # ^  is equivalent to symmetric_difference
    # *  is equivalent to cross
    
    __le__  = issubset
    __ge__  = issuperset
    __or__  = union
    __and__ = intersection
    __sub__ = difference
    __xor__ = symmetric_difference
    __mul__ = cross

    def __lt__(self,other):
        """
        Return True if the set is a strict subset of 'other'

        TODO: verify that this is more efficient than an explicit implimentation.
        """
        return self <= other and not self == other

    def __gt__(self,other):
        """
        Return True if the set is a strict superset of 'other'

        TODO: verify that this is more efficient than an explicit implimentation.
        """
        return self >= other and not self == other

    def construct(self, values=None):
        """
        Apply the rule to construct values in this set

        TODO: rework to avoid redundant code
        """
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing SimpleSet, name="+self.cname(True)+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True

        if self.initialize is None:                             # TODO: deprecate this functionality
            self.initialize = getattr(self,'rule',None)
            if not self.initialize is None:
                logger.warning("DEPRECATED: The set 'rule' attribute cannot be used to initialize component "+self.cname(True)+".  Use the 'initialize' attribute")
        #
        # Construct using the input values list
        #
        if values is not None:
            if type(self._bounds) is tuple:
                first=self._bounds[0]
                last=self._bounds[1]
            else:
                first=None
                last=None
            all_numeric=True
            #
            # TODO: verify that values is not a list
            #
            for val in values[None]:
                #
                # Skip the value if it is filtered
                #
                if not self.filter is None and not apply_indexed_rule(self, self.filter, self._parent(), val):
                    continue
                self.add(val)
                if type(val) in intlist:
                    if first is None or val<first:
                        first=val
                    if last is None or val>last:
                        last=val
                else:
                    all_numeric=False
            if all_numeric:
                self._bounds = (first, last)
        #
        # Construct using the initialize rule
        #
        elif type(self.initialize) is types.FunctionType:
            if self._parent is None:
                raise ValueError("Must pass the parent block in to initialize with a function")
            if self.initialize.__code__.co_argcount == 1:
                #
                # Using a rule of the form f(model) -> iterator
                #
                tmp = self.initialize(self._parent())
                for val in tmp:
                    if self.dimen == 0:
                        if type(val) in [tuple,list]:
                            self.dimen=len(val)
                        else:
                            self.dimen=1
                    if not self.filter is None and \
                       not apply_indexed_rule(self, self.filter, self._parent(), val):
                        continue
                    self.add(val)
            else:
                #
                # Using a rule of the form f(model, z) -> element
                #
                ctr=1
                val = apply_indexed_rule(self, self.initialize, self._parent(), ctr)
                if val is None:
                    raise ValueError("Set rule returned None instead of Set.Skip")
                if self.dimen == 0:
                    if type(val) in [tuple,list] and not val == Set.End:
                        self.dimen=len(val)
                    else:
                        self.dimen=1
                while not (val.__class__ is tuple and val == Set.End):
                    # Add the value if the filter is None or the filter return value is True
                    if self.filter is None or \
                       apply_indexed_rule(self, self.filter, self._parent(), val):
                       self.add(val)
                    ctr += 1
                    val = apply_indexed_rule(self, self.initialize, self._parent(), ctr)
                    if val is None:
                        raise ValueError("Set rule returned None instead of Set.Skip")
        
            # Update the bounds if after using the rule, the set is 
            # a one dimensional list of all numeric values
            if self.dimen == 1:
                if type(self._bounds) is tuple:
                    first=self._bounds[0]
                    last=self._bounds[1]
                else:
                    first=None
                    last=None
                all_numeric=True
                for val in self.value:
                    if type(val) in intlist:
                        if first is None or val<first:
                            first=val
                        if last is None or val>last:
                            last=val
                    else:
                        all_numeric=False
                        break
                if all_numeric:
                    self._bounds = (first, last)

        #
        # Construct using the default values
        #
        elif self.initialize is not None:
            if type(self.initialize) is dict:
                raise ValueError("Cannot initialize set "+self.cname(True)+" with dictionary data")
            if type(self._bounds) is tuple:
                first=self._bounds[0]
                last=self._bounds[1]
            else:
                first=None
                last=None
            all_numeric=True
            for val in self.initialize:
                # Skip the value if it is filtered
                if not self.filter is None and \
                   not apply_indexed_rule(self, self.filter, self._parent(), val):
                    continue
                if type(val) in intlist:
                    if first is None or val<first:
                        first=val
                    if last is None or val>last:
                        last=val
                else:
                    all_numeric=False
                self.add(val)
            if all_numeric:
                self._bounds = (first,last)


class SimpleSet(SimpleSetBase,_SetData):

    def __init__(self, *args, **kwds):
        self._bounds = kwds.pop('bounds', None)
        SimpleSetBase.__init__(self, *args, **kwds)
        _SetData.__init__(self, self, self._bounds)

    def __getitem__(self, key):
        """
        Return the specified member of the set.

        This method generates an exception because the set is unordered.
        """
        return _SetData.__getitem__(self, key)


class OrderedSimpleSet(SimpleSetBase,_OrderedSetData):

    def __init__(self, *args, **kwds):
        self._bounds = kwds.pop('bounds', None)
        SimpleSetBase.__init__(self, *args, **kwds)
        _OrderedSetData.__init__(self, self, self._bounds)

    def __getitem__(self, key):
        """
        Return the specified member of the set.
        """
        return _OrderedSetData.__getitem__(self, key)


# REVIEW - START

class SetOf(SimpleSet):
    """
    A derived SimpleSet object that creates a set from external
    data without duplicating it.
    """

    def __init__(self, *args, **kwds):
        if len(args) != 1:
            raise TypeError("Only one set data argument can be specified")
        self.dimen = 0
        SimpleSet.__init__(self,**kwds)
        self._elements = args[0]
        self.value = None
        self._constructed = True
        self._bounds = (None, None) # We cannot determine bounds, since the data may change
        self.virtual = False
        try:
            len(args[0])
            self.concrete = True
        except:
            self.concrete = False
        #
        if self.dimen == 0:
            try:
                for i in self._elements:
                    if type(i) is tuple:
                        self.dimen = len(i)
                    else:
                        self.dimen = 1
                    break
            except TypeError:
                e = sys.exc_info()[1]
                raise TypeError("Cannot create a Pyomo set: "+e)

    def construct(self, values=None):
        """
        Disabled construction method
        """
        pass

    def __len__(self):
        """
        The number of items in the set.
        """
        try:
            return len(self._elements)
        except:
            pass
        #
        # If self._elements cannot provide size information,
        # then we need to iterate through all set members.
        #
        ctr = 0
        for i in self:
            ctr += 1
        return ctr

    def __iter__(self):
        """
        Return an iterator for the underlying set
        """
        for i in self._elements:
            yield i

    def _set_contains(self, element):
        """
        A wrapper function that tests if the element is in
        the data associated with a concrete set.
        """ 
        return element in self._elements
        
    def data(self):
        """
        Return the underlying set data by constructing
        a python set() object explicitly.
        """
        return set(self)

    def __eq__(self, other):
        """
        Equality comparison

	    Note that this checks the length first only if the
        current set is concrete.  This is an optimization that
        may avoid iterating over all set elements.
        """
        if other is None:
            return False
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            return False
        if other.concrete and len(self) != len(other):
            return False
        ctr = 0
        for i in other:
            if not i in self:
                return False
            ctr += 1
        return other.concrete or ctr == len(self)
            

class _SetOperator(SimpleSet):
    """A derived SimpleSet object that contains a concrete virtual single set."""

    def __init__(self, *args, **kwds):
        if len(args) != 2:
            raise TypeError("Two arguments required for a binary set operator")
        dimen_test = kwds.get('dimen_test',True)
        if 'dimen_test' in kwds:
            del kwds['dimen_test']
        SimpleSet.__init__(self,**kwds)
        self.value = None
        self._constructed = True
        self.virtual = True
        self.concrete = True
        #
        self._setA = args[0]
        if not self._setA.concrete:
            raise TypeError("Cannot perform set operations with non-concrete set '"+self._setA.cname(True)+"'")
        if isinstance(args[1],SimpleSet):
            self._setB = args[1]
        else:
            self._setB = SetOf(args[1])
        if not self._setB.concrete:
            raise TypeError("Cannot perform set operations with non-concrete set '"+self._setB.cname(True)+"'")
        if dimen_test and self._setA.dimen != self._setB.dimen:
            raise ValueError("Cannot perform set operation with sets "+self._setA.cname(True)+" and "+self._setB.cname(True)+" that have different element dimensions: "+str(self._setA.dimen)+" "+str(self._setB.dimen))
        self.dimen = self._setA.dimen
        #
        self.ordered = self._setA.ordered and self._setB.ordered

    def construct(self, values=None):
        """ Disabled construction method """
        pass

    def __len__(self):
        """The number of items in the set."""
        ctr = 0
        for i in self:
            ctr += 1
        return ctr

    def __iter__(self):
        """Return an iterator for the underlying set"""
        raise IOError("Undefined set iterator")

    def _set_contains(self, element):
        raise IOError("Undefined set operation")
        
    def data(self):
        """The underlying set data."""
        return set(self)

    def __eq__(self, other):
        """ Equality comparison """
        if other is None:
            return False
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            return False
        if other.concrete and len(self) != len(other):
            return False
        ctr = 0
        for i in self:
            if not i in other:
                return False
            ctr += 1
        return other.concrete or ctr == len(other)
            

class _SetUnion(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            yield elt
        for elt in self._setB:
            if not elt in self._setA:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA or elt in self._setB
    

class _SetIntersection(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            if elt in self._setB:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA and elt in self._setB
    

class _SetDifference(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            if not elt in self._setB:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA and not elt in self._setB
    

class _SetSymmetricDifference(_SetOperator):

    def __init__(self, *args, **kwds):
        _SetOperator.__init__(self, *args, **kwds)

    def __iter__(self):
        for elt in self._setA:
            if not elt in self._setB:
                yield elt
        for elt in self._setB:
            if not elt in self._setA:
                yield elt

    def _set_contains(self, elt):
        return elt in self._setA ^ elt in self._setB
    

class _SetProduct(_SetOperator):

    def __init__(self, *args, **kwd):
        kwd['dimen_test'] = False
        
        # every input argument in a set product must be iterable. 
        for arg in args:
            try:
                iter(arg)
            except TypeError:
                raise TypeError("Each input argument to a _SetProduct constructor must be iterable")

        _SetOperator.__init__(self, *args, **kwd)
        # the individual index sets definining the product set.
        if isinstance(self._setA,_SetProduct):
            self.set_tuple = self._setA.set_tuple
        else:
            self.set_tuple = [self._setA]
        if isinstance(self._setB,_SetProduct):
            self.set_tuple += self._setB.set_tuple
        else:
            self.set_tuple.append(self._setB)
        # set the "dimen" instance attribute.
        self._compute_dimen()

    def __iter__(self):
        if self.is_flat_product():
            for i in itertools.product(*self.set_tuple):
                yield i
        else:
            for i in itertools.product(*self.set_tuple):
                yield pyutilib_misc_flatten_tuple(i)

    def _set_contains(self, element):
        # Do we really need to check if element is a tuple???
        # if type(element) is not tuple:
        #    return False
        try:
            ctr = 0
            for subset in self.set_tuple:
                d = subset.dimen
                if d == 1:
                    if not subset._set_contains(element[ctr]):
                        return False
                else:
                    # cast to tuple is not needed: slices of tuples
                    # return tuples!
                    if not subset._set_contains(element[ctr:ctr+d]):
                        return False

                ctr += d
            return ctr == len(element)
        except:
            return False
        
    def __len__(self):
        ans = 1
        for _set in self.set_tuple:
            ans *= len(_set)
        return ans

    def _compute_dimen(self):
        ans=0
        for _set in self.set_tuple:
            if _set.dimen is None:
                self.dimen=None
                return
            else:
                ans += _set.dimen
        self.dimen = ans

    def is_flat_product(self):
        """
        a simple utility to determine if each of the composite sets is
        of dimension one. knowing this can significantly reduce the
        cost of iteration, as you don't have to call flatten_tuple.
        """

        for s in self.set_tuple:
            if s.dimen != 1:
                return False
        return True

    def _verify(self, element):
        """
        If this set is virtual, then an additional check is made
        to ensure that the element is in each of the underlying sets.
        """
        tmp = SimpleSet._verify(self, element)
        return tmp

        # WEH - when is this needed?
        if not tmp or not self.virtual:
            return tmp

        next_tuple_index = 0
        member_set_index = 0
        for member_set in self.set_tuple:
            tuple_slice = element[next_tuple_index:next_tuple_index + member_set.dimen]
            if member_set.dimen == 1:
                tuple_slice = tuple_slice[0]
            if tuple_slice not in member_set:
                return False
            member_set_index += 1
            next_tuple_index += member_set.dimen
        return True

# REVIEW - END

class IndexedSet(Set):
    """
    An array of sets, which are indexed by other sets
    """

    def __init__(self, *args, **kwds):      #pragma:nocover
        self._bounds = kwds.pop("bounds", None)
        Set.__init__(self, *args, **kwds)
        if 'virtual' in kwds:                                       #pragma:nocover
            raise TypeError("It doesn't make sense to create a virtual set array")
        if self.ordered:
            self._SetData = _IndexedOrderedSetData
        else:
            self._SetData = _IndexedSetData

    def size(self):
        """
        Return the number of elements in all of the indexed sets.
        """
        ans = 0
        for cdata in itervalues(self):
            ans += len(cdata)
        return ans

    def data(self):
        """
        Return the dictionary of sets
        """
        return self._data

    def clear(self):
        """
        Clear that data in this component.
        """
        if UnindexedComponent_set != self._index:
            self._data = {}
        else:
            #
            # TODO: verify that this could happen
            #
            pass

    def _default(self, index):
        """
        Return the default component data value

        This returns an exception.
        """
        raise KeyError(
            "Index %s not defined for set array %s" % (index, self.__class__.__name__,))

    def __setitem__(self, key, vals):
        """
        Add a set to the index.
        """
        if key not in self._index:
            raise KeyError("Cannot set index "+str(key)+" in array set "+self.cname(True))
        #
        # Create a _SetData object if one doesn't already exist
        #
        if key in self._data:
            self._data[key].value.clear()
        else:
            self._data[key] = self._SetData(self, self._bounds)
        #
        # Add the elements in vals to the _SetData object
        #
        _set = self._data[key]
        for elt in vals:
            _set.add(elt)

    def check_values(self):
        """
        Verify the values of all indexed sets.

        TODO: document when unverified values could be set.
        """
        for cdata in itervalues(self):
            for val in cdata.value:
                self._verify(val)

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        _ordered = self.ordered
        if type(_ordered) is bool:
            pass
        elif _ordered is Set.InsertionOrder:
            _ordered = 'Insertion'
        elif _ordered is Set.SortedOrder:
            _ordered = 'Sorted'
        else:
            _ordered = '{user}'
        return (
            [("Dim", self.dim()),
             ("Dimen", self.dimen),
             ("Size", self.size()),
             ("Domain", None if self.domain is None else self.domain.name),
             ("ArraySize", len(self._data)),
             ("Ordered", _ordered),
             ("Bounds", self._bounds)],
            iteritems(self._data),
            ("Key","Members"),
            lambda k, v: [ 
                k,
                v.value if self.ordered else sorted(v.value) ] )
            
    def construct(self, values=None):
        """
        Apply the rule to construct values in each set
        """
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing IndexedSet, name="+self.cname(True)+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True
        #
        if self.initialize is None:             # TODO: deprecate this functionality
            self.initialize = getattr(self,'rule',None)
            if not self.initialize is None:
                logger.warning("DEPRECATED: The set 'rule' attribute cannot be used to initialize component "+self.cname(True)+".  Use the 'initialize' attribute")
        #
        # Construct using the values dictionary
        #
        if values is not None:
            for key in values:
                if type(key) is tuple and len(key)==1:
                    tmpkey=key[0]
                else:
                    tmpkey=key
                if tmpkey not in self._index:
                    raise KeyError("Cannot construct index "+str(tmpkey)+" in array set "+self.cname(True))
                tmp = self._SetData(self, self._bounds)
                for val in values[key]:
                    tmp._add(val)
                self._data[tmpkey] = tmp
        #
        # Construct using the rule
        #
        elif type(self.initialize) is types.FunctionType:
            if self._parent is None:
                raise ValueError("Need parent block to construct a set array with a function")
            for key in self._index:
                tmp = self._SetData(self, self._bounds)
                self._data[key] = tmp
                #
                if isinstance(key,tuple):
                    tmpkey = key
                else:
                    tmpkey = (key,)
                #
                # self.initialize: model, index -> list
                #
                if self.initialize.__code__.co_argcount == len(tmpkey)+1:
                    rule_list = apply_indexed_rule(self, self.initialize, self._parent(), tmpkey)
                    for val in rule_list:
                        tmp._add( val )
                #
                # self.initialize: model, counter, index -> val
                #
                else:
                    ctr=1
                    val = apply_parameterized_indexed_rule(self, self.initialize, self._parent(), ctr, tmpkey)
                    if val is None:
                        raise ValueError("Set rule returned None instead of Set.Skip")
                    while not (val.__class__ is tuple and val == Set.End):
                        tmp._add( val )
                        ctr += 1
                        val = apply_parameterized_indexed_rule(self, self.initialize, self._parent(), ctr, tmpkey)
                        if val is None:
                            raise ValueError("Set rule returned None instead of Set.Skip")
        #
        # Treat self.initialize as an iterable
        #
        elif self.initialize is not None:
            if type(self.initialize) is not dict:
                for key in self._index:
                    tmp = self._SetData(self, self._bounds)
                    for val in self.initialize:
                        tmp._add(val)
                    self._data[key] = tmp
            else:
                for key in self.initialize:
                    tmp = self._SetData(self, self._bounds)
                    for val in self.initialize[key]:
                        tmp._add(val)
                    self._data[key] = tmp



register_component(SetOf, "Define a Pyomo Set component using an iterable data object.")
register_component(Set, "Set data that is used to define a model instance.")




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

from __future__ import division

__all__ = [ 'generate_canonical_repn', 'as_expr', 'canonical_is_constant', 
            'canonical_is_linear', 'canonical_is_quadratic', 'canonical_is_nonlinear', 
            'canonical_degree', 'LinearCanonicalRepn', 'GeneralCanonicalRepn' ]

import logging
import copy
import six

from six import iterkeys, itervalues, iteritems, StringIO
from six.moves import xrange, reduce

using_py3 = six.PY3

from coopr.pyomo.base import IPyomoPresolver, IPyomoPresolveAction, Model, \
                             Constraint, Objective, value
from coopr.pyomo.base import param
from coopr.pyomo.base import expr
from coopr.pyomo.base.expression import _ExpressionData, SimpleExpression, Expression
from coopr.pyomo.base.connector import _ConnectorValue, SimpleConnector
from coopr.pyomo.base.var import _VarDataWithDomain, SimpleVar, Var, _VarData

logger = logging.getLogger('coopr.pyomo')

##############################################################################
##############################################################################
#### CLASSES AND ROUTINES FOR GENERAL CANONICAL REPRESENTATIONS START NOW ####
##############################################################################
##############################################################################

#
# A frozen dictionary that can be hashed.  This dictionary isn't _really_
# frozen, but it acts hashable.
#
class GeneralCanonicalRepn(dict):

    __slots__ = ['_hash']

    def __hash__(self):
        rval = getattr(self, '_hash', None)
        if rval is None:
            rval = self._hash = hash(frozenset(iteritems(self)))
        return rval

    # getstate and setstate are currently not defined in a generic manner, i.e.,
    # we assume the underlying dictionary is the only thing

    def __getstate__(self):
        return self.items()

    def __setstate__(self, dictionary):
        self.clear()
        self.update(dictionary)
        self._hash = None

    def __str__(self):
        return "GeneralCanonical{ %s%s }" % (
            as_expr(self, ignore_other=True),
            ", general: %s" % self[None] if None in self else "" )

#
# Generate a canonical representation of an expression.
#
# The canonical representation is a dictionary.  Each element is a mapping
# from a term degree to terms in the expression.  If the term degree is
# None, then the map value is simply an expression of terms without a
# specific degree.  Otherwise, the map value is a dictionary of terms to
# their coefficients.  A term is represented as a frozen dictionary that
# maps variable id to variable power.  A constant term is represented
# with None.
#
# Examples:
#  Let x[1] ... x[4] be the first 4 variables, and
#      y[1] ... y[4] be the next 4 variables
#
# 1.3                           {0:{ None :1.3}}
# 3.2*x[1]                      {1:{ {0:1} :3.2}}
# 2*x[1]*y[2] + 3*x[2] + 4      {0:{None:4.0}, 1:{{1:1}:3.0}, 2:{{0:1, 5:1}:2.0}}
# 2*x[1]**4*y[2]    + 4         {0:{None:4.0}, 1:{{1:1}:3.0}, 5:{{0:4, 5:1 }:2.0}}
# log(y[1]) + x[1]*x[1]         {2:{{0:2}:1.0}, None:log(y[1])}
#



def as_expr(rep, vars=None, model=None, ignore_other=False):
    """ Convert a canonical representation into an expression. """
    if isinstance(model, Model):
        vars = model._var
        id_offset=0
    elif not vars is None:
        id_offset=1
    exp = 0.0
    for d in sorted(rep):
        if d in (None, -1):
            continue
        for v in sorted(rep[d]):
            if v is None:
                exp += rep[d][v]
                continue
            e = rep[d][v]
            for id in v:
                if vars is None:
                    e *= rep[-1][id]**v[id]
                else:
                    e *= vars[id[1]+id_offset]**v[id]
            exp += e
    # Stick general nonlineat at the end
    if None in rep and not ignore_other:
        exp += rep[None]
    return exp

def repn_add(lhs, rhs, coef=1.0):
    """
    lhs and rhs are the expressions being added together.
    'lhs' and 'rhs' are left-hand and right-hand side operands
    See generate_canonical_repn for explanation of pyomo expressions
    """
    for order in rhs:
        # For each order term, first-order might be 3*x,
        # second order 4*x*y or 5*x**2
        if order is None:
            # i.e., (currently) order is a constant or logarithm
            if order in lhs:
                lhs[order] += rhs[order]
            else:
                lhs[order] = rhs[order]
            continue
        if order < 0:
            # ignore now, handled below
            continue
        if not order in lhs:
            lhs[order] = {}
        for var in rhs[order]:
            # Add coefficients of variables in this order (e.g., third power)
            lhs[order][var] = coef*rhs[order][var] + lhs[order].get(var,0.0)
    #
    # Merge the _VarData maps
    #
    if -1 in rhs:
        if -1 in lhs:
            lhs[-1].update(rhs[-1])
        else:
            lhs[-1] = rhs[-1]
    return lhs

def repn_mult(r1, r2, coef=1.0):
    rep = {}
    for d1 in r1:
        for d2 in r2:
            if d1 == None or d2 == None or d1 < 0 or d2 < 0:
                pass
            else:
                d=d1+d2
                if not d in rep:
                    rep[d] = {}
                if d == 0:
                    rep[d][None] = coef * r1[0][None] * r2[0][None]
                elif d1 == 0:
                    for v2 in r2[d2]:
                        rep[d][v2]  = coef * r1[0][None] * r2[d2][v2] + rep[d].get(v2,0.0)
                elif d2 == 0:
                    for v1 in r1[d1]:
                        rep[d][v1]  = coef * r1[d1][v1] * r2[0][None] + rep[d].get(v1,0.0)
                else:
                    for v1 in r1[d1]:
                        for v2 in r2[d2]:
                            v = GeneralCanonicalRepn(v1)
                            for id in v2:
                                if id in v:
                                    v[id] += v2[id]
                                else:
                                    v[id]  = v2[id]
                            rep[d][v] = coef * r1[d1][v1] * r2[d2][v2] + rep[d].get(v,0.0)
    #
    # Handle other nonlinear terms
    #
    if None in r1:
        rep[None] = as_expr(r2, ignore_other=True) * copy.deepcopy(r1[None])
        if None in r2:
            rep[None] += copy.deepcopy(r1[None])*copy.deepcopy(r2[None])
    if None in r2:
        if None in rep:
            rep[None] += as_expr(r1, ignore_other=True) * copy.deepcopy(r2[None])
        else:
            rep[None]  = as_expr(r1, ignore_other=True) * copy.deepcopy(r2[None])
    #
    # Merge the _VarData maps
    #
    if -1 in r1:
        rep[-1] = r1[-1]
    if -1 in r2:
        if -1 in rep:
            rep[-1].update(r2[-1])
        else:
            rep[-1] = r2[-1]
    #
    # Return the canonical repn
    #
    return rep

def collect_variables(exp, idMap):
    if exp.is_expression():
        ans = {}
        if exp.__class__ is expr._ProductExpression:
            for subexp in exp._numerator:
                ans.update(collect_variables(subexp, idMap))
            for subexp in exp._denominator:
                ans.update(collect_variables(subexp, idMap))
        else:
            # This is fragile: we assume that all other expression
            # objects "play nice" and just use the _args member.
            for subexp in exp._args:
                ans.update(collect_variables(subexp, idMap))
        return ans
    elif exp.is_fixed():
        # NB: is_fixed() returns True for constants and variables with
        # fixed values
        return {}
    elif exp.__class__ is _VarData or exp.__class__ is _VarDataWithDomain or exp.type() is Var:
        id_ = id(exp)
        if id_ in idMap[None]:
            key = idMap[None][id_]
        else:
            key = len(idMap) - 1
            idMap[None][id_] = key
            idMap[key] = exp
        return { key : exp }
    else:
        raise ValueError("Unexpected expression type: "+str(exp))

#
# Temporary canonical expressions
#
# temp_const = { 0: {None:0.0} }
# temp_var = { 1: {GeneralCanonicalRepn({None:1}):1.0} }
# temp_nonl = { None: None }

#
# Internal function for collecting canonical representation, which is
# called recursively.
#
def collect_general_canonical_repn(exp, idMap, compute_values):
#     global temp_const
#     global temp_var
#     global temp_nonl
    temp_const = { 0: {None:0.0} }
    temp_var = { 1: {GeneralCanonicalRepn({None:1}):1.0} }
    temp_nonl = { None: None }
    exp_type = type(exp)

    #
    # Constant
    #
    if exp.is_fixed():
        if compute_values:
            temp_const[0][None] = value(exp)
        else:
            temp_const[0][None] = exp
        return temp_const
    #
    # Expression
    #
    elif exp.is_expression():

        #
        # Sum
        #
        if exp_type is expr._SumExpression:
            if exp._const != 0.0:
                repn = { 0: {None:exp._const} }
            else:
                repn = {}
            for i in xrange(len(exp._args)):
                repn = repn_add(repn, collect_general_canonical_repn(exp._args[i], idMap, compute_values), coef=exp._coef[i] )
            return repn
        #
        # Product
        #
        elif exp_type is expr._ProductExpression:
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.is_fixed():
                    denom *= e()
                else:
                    temp_nonl[None] = exp
                    return temp_nonl
                if denom == 0.0:
                    print("Divide-by-zero error - offending sub-expression:")
                    e.pprint()
                    raise ZeroDivisionError
            #
            # OK, the denominator is a constant.
            #
            repn = { 0: {None:exp._coef / denom} }
            for e in exp._numerator:
                repn = repn_mult(repn, collect_general_canonical_repn(e, idMap, compute_values))
            return repn
        #
        # Power Expression
        #
        elif exp_type is expr._PowExpression:
            if exp.polynomial_degree() is None:
                raise TypeError("Unsupported general power expression: "+str(exp._args))
                
            # If this is of the form EXPR**1, we can just get the
            # representation of EXPR
            if exp._args[1] == 1:
                return collect_general_canonical_repn(exp._args[0], idMap, compute_values)
            # The only other way to get a polynomial expression is if
            # exp=EXPR**p where p is fixed a nonnegative integer.  We
            # can expand this expression and generate a canonical
            # representation from there.  If p=0, this expression is
            # constant (and is processed by the is_fixed code above
            # NOTE: There is no check for 0**0
            return collect_general_canonical_repn(
                        reduce( lambda x,y: x*y, [exp._args[0]]*int(exp._args[1]), 1.0 ),
                        idMap,
                        compute_values)
        elif exp_type is expr.Expr_if:
            if exp._if.is_fixed():
                if exp._if():
                    return collect_general_canonical_repn(exp._then, idMap, compute_values)
                else:
                    return collect_general_canonical_repn(exp._else, idMap, compute_values)
            else:
                temp_nonl[None] = exp
                return temp_nonl
        #
        # Expression (the component)
        # 
        elif exp_type is _ExpressionData or exp.type() is Expression:
            return collect_general_canonical_repn(exp.value, idMap, compute_values)
        #
        # ERROR
        #
        else:
            raise ValueError("Unsupported expression type: "+str(exp))
    #
    # Variable
    #
    elif exp_type is _VarData or exp.__class__ is _VarDataWithDomain or exp.type() is Var:
        id_ = id(exp)
        if id_ in idMap[None]:
            key = idMap[None][id_]
        else:
            key = len(idMap) - 1
            idMap[None][id_] = key
            idMap[key] = exp
        temp_var = { -1: {key:exp}, 1: {GeneralCanonicalRepn({key:1}):1.0} }
        return temp_var
    #
    # Connector
    #
    elif exp_type is _ConnectorValue or exp.type() is Connector:
        # Silently omit constraint...  The ConnectorExpander should
        # expand this constraint into indvidual constraints that
        # reference "real" variables.
        return {}
    #
    # ERROR
    #
    else:
        raise ValueError("Unexpected expression (type %s): " %
                         ( type(exp).__name__, str(exp) ))

##############################################################################
##############################################################################
#### CLASSES AND ROUTINES FOR LINEAR CANONICAL REPRESENTATIONS START NOW #####
##############################################################################
##############################################################################

# not a named tuple, because we want the fields mutable.
class LinearCanonicalRepn(object):

    __slots__ = ['variables', 'constant', 'linear']

    def __init__(self, **kwds):

        # a single scalar (int or float)
        self.constant = None 

        # a tuple of coefficents - one-for-one with self.variables
        self.linear = None    

        # a tuple of _VarValues
        self.variables = None

    def __getstate__(self):
        """
        This method is required because this class uses slots.
        """
        return dict((i, getattr(self, i, None)) for i in LinearCanonicalRepn.__slots__)

    def __setstate__(self, state):
        """
        This method is required because this class uses slots.
        """
        for (slot_name, value) in iteritems(state):
            setattr(self, slot_name, value)

    def __str__(self):
        ordered_vars = None
        if self.variables is not None:
            ordered_vars = sorted( 
                (v.cname(True), i) for i,v in enumerate(self.variables) )
        tmp_str = str(self.constant) if (self.constant) else ("")
        tmp_str += (" + ") if (self.constant and self.variables) else ("")
        tmp_str += (" + ".join("%s*%s"%(self.linear[i], v) for v,i in ordered_vars)) if (self.variables) else ("")
        return "LinearCanonical{ %s }" % (tmp_str)

def _collect_linear_sum(exp, idMap, multiplier, coef, varmap, compute_values):

    coef[None] += multiplier * exp._const  # None is the constant term in the coefficient map.
    
    arg_coef_iterator = exp._coef.__iter__()
    for arg in exp._args:
        # an arg can be anything - a product, a variable, whatever.

        # Special case... <sigh>
        if (arg.__class__ is _VarData or arg.__class__ is _VarDataWithDomain) and arg.fixed is False:
            # save an expensive recursion - this is by far the most common case.
            id_ = id(arg)
            if id_ in idMap[None]:
                key = idMap[None][id_]
            else:
                key = len(idMap) - 1
                idMap[None][id_] = key
                idMap[key] = arg
            #
            varmap[key]=arg
            if key in coef:
                coef[key] += multiplier * six.next(arg_coef_iterator)
            else:
                coef[key] = multiplier * six.next(arg_coef_iterator)
        else:
            _linear_collectors[arg.__class__](arg, idMap, multiplier * six.next(arg_coef_iterator), coef, varmap, compute_values)

def _collect_linear_prod(exp, idMap, multiplier, coef, varmap, compute_values):

    multiplier *= exp._coef
    _coef = { None : 0 }
    _varmap = {}
    
    for subexp in exp._denominator:
        if compute_values:
            x = value(subexp) # only have constants/fixed terms in the denominator.
            if x == 0:
                buf = StringIO()
                subexp.pprint(buf)
                logger.error("Divide-by-zero: offending sub-expression:\n   " + buf)
                raise ZeroDivisionError
            multiplier /= x
        else:
            multiplier /= subexp

    for subexp in exp._numerator:
        if _varmap:
            if compute_values:
                multiplier *= value(subexp)
            else:
                multiplier *= subexp
        else:
            _linear_collectors[subexp.__class__](subexp, idMap, 1, _coef, _varmap, compute_values)
            if not _varmap:
                multiplier *= _coef[None]
                _coef[None] = 0

    if _varmap:
        for key, val in iteritems(_coef):
            if key in coef:
                coef[key] += multiplier * val
            else:
                coef[key] = multiplier * val
        varmap.update(_varmap)
    else:
        # constant expression; i.e. 1/x
        coef[None] += multiplier
        

def _collect_linear_pow(exp, idMap, multiplier, coef, varmap, compute_values):

    if exp.is_fixed():
        if compute_values:
            coef[None] += multiplier * value(exp)
        else:
            coef[None] += multiplier * exp
    elif value(exp._args[1]) == 1:
        arg = exp._args[0]
        _linear_collectors[arg.__class__](arg, idMap, multiplier, coef, varmap, compute_values)
    else:
        raise TypeError( "Unsupported power expression: "+str(exp._args) )

def _collect_branching_expr(exp, idMap, multiplier, coef, varmap, compute_values):

    if exp._if.is_fixed():
        if exp._if():
            arg = exp._then
            _linear_collectors[arg.__class__](arg, idMap, multiplier, coef, varmap, compute_values)
        else:
            arg = exp._else
            _linear_collectors[arg.__class__](arg, idMap, multiplier, coef, varmap, compute_values)
    else:
        raise TypeError( "Unsupported dynamic If-Then-Else expression: "+str(exp._args) )

def _collect_linear_var(exp, idMap, multiplier, coef, varmap, compute_values):

    if exp.is_fixed():
        if compute_values:
            coef[None] += multiplier * value(exp)
        else:
            coef[None] += multiplier * exp
    else:
        id_ = id(exp)
        if id_ in idMap[None]:
            key = idMap[None][id_]
        else:
            key = len(idMap) - 1
            idMap[None][id_] = key
            idMap[key] = exp
        #
        if key in coef:
            coef[key] += multiplier
        else:
            coef[key] = multiplier
        varmap[key] = exp

def _collect_linear_const(exp, idMap, multiplier, coef, varmap, compute_values):
    if compute_values:
        coef[None] += multiplier * value(exp)
    else:
        coef[None] += multiplier * exp

def _collect_linear_connector(exp, idMap, multiplier, coef, varmap, compute_values):

    pass

def _collect_linear_intrinsic(exp, idMap, multiplier, coef, varmap, compute_values):

    if exp.is_fixed():
        if compute_values:
            coef[None] += multiplier * value(exp)
        else:
            coef[None] += multiplier * exp
    else:
        raise TypeError( "Unsupported intrinsic expression: %s: %s" % (exp, str(exp._args)) )

def _collect_identity(exp, idMap, multiplier, coef, varmap, compute_values):
    exp = exp.value
    if exp.is_fixed():
        if compute_values:
            coef[None] += multiplier * value(exp)
        else:
            coef[None] += multiplier * exp
    else:
        _linear_collectors[exp.__class__](exp, idMap, multiplier, coef, varmap, compute_values)


_linear_collectors = {
    expr._SumExpression               : _collect_linear_sum,
    expr._ProductExpression           : _collect_linear_prod,
    expr._PowExpression               : _collect_linear_pow,
    expr._IntrinsicFunctionExpression : _collect_linear_intrinsic,
    _ConnectorValue         : _collect_linear_connector,
    SimpleConnector         : _collect_linear_connector,
    expr.Expr_if            : _collect_branching_expr,
    param._ParamData        : _collect_linear_const,
    param.SimpleParam       : _collect_linear_const,
    param.Param             : _collect_linear_const,
    _VarData                : _collect_linear_var,
    _VarDataWithDomain      : _collect_linear_var,
    SimpleVar               : _collect_linear_var,
    Var                     : _collect_linear_var,
    _ExpressionData         : _collect_identity,
    SimpleExpression        : _collect_identity,
    Expression              : _collect_identity
    }

def collect_linear_canonical_repn(exp, idMap, compute_values=True):

    idMap.setdefault(None, {})
    coef = { None : 0 }
    varmap = {}
    try:
        _linear_collectors[exp.__class__](exp, idMap, 1, coef, varmap, compute_values)
    except KeyError:
        raise
        raise ValueError( "Unexpected expression (type %s): %s" %
                          (type(exp).__name__, str(exp)) )
    return coef, varmap

#########################################################################
#########################################################################
#### ROUTINES OPERATING ON BOTH LINEAR AND GENERAL CANONICAL REPNS  #####
#########################################################################
#########################################################################

def generate_canonical_repn(exp, idMap=None, compute_values=True):

    degree = exp.polynomial_degree()

    if idMap is None:
        idMap = {}
    idMap.setdefault(None, {})

    if degree == 0:
        ans = LinearCanonicalRepn()
        ans.constant = value(exp)
        return ans

    elif degree == 1:
        # varmap is a map from the variable id() to a _VarData.
        # coef is a map from the variable id() to its coefficient.
        coef, varmap = collect_linear_canonical_repn(exp, idMap, compute_values)
        ans = LinearCanonicalRepn()
        if None in coef:
            val = coef.pop(None)
            if val != 0.0:
                ans.constant = val

        # the six module is inefficient in terms of wrapping iterkeys
        # and itervalues, in the context of Python 2.7. use the native
        # dictionary methods where possible.
        if using_py3:
            ans.linear = tuple( itervalues(coef) )
            ans.variables = tuple(varmap[var_hash] for var_hash in iterkeys(coef) )
        else:
            ans.linear = tuple( coef.itervalues() )
            ans.variables = tuple(varmap[var_hash] for var_hash in coef.iterkeys() )
        return ans

    # **Py3k: degree > 1 comparision will error if degree is None
    elif degree and degree > 1:
        ans = collect_general_canonical_repn(exp, idMap, compute_values)
        if 1 in ans:
            linear_terms = {}
            for key, coef in iteritems(ans[1]):
                linear_terms[list(key.keys())[0]] = coef
            ans[1] = linear_terms
        return GeneralCanonicalRepn(ans)
    else:
        return GeneralCanonicalRepn(
            { None: exp, -1 : collect_variables(exp, idMap) } )

def canonical_is_constant(repn):
    """Return True if the canonical representation is a constant expression"""
    if isinstance(repn, dict):
        return max(iterkeys(repn)) == 0 and None not in repn
    else:
        return (repn.constant is not None) and (repn.linear is None)

def canonical_is_nonlinear(repn):
    """Return True if the canonical representation is a nonlinear expression"""
    if isinstance(repn, dict):
        return max(iterkeys(repn)) > 1 or None in repn
    else:
        return False

def canonical_is_linear(repn):
    """Return True if the canonical representation is a linear expression."""
    if isinstance(repn, dict):
        return max(iterkeys(repn)) == 1 and None not in repn
    else:
        return (repn.linear is not None)

def canonical_is_quadratic(repn):
    """Return True if the canonical representation is a quadratic expression."""
    if isinstance(repn, dict):
        return max(iterkeys(repn)) == 2 and None not in repn
    else:
        return False

def canonical_degree(repn):
    if isinstance(repn, dict):
        if None in repn:
            return None
        else:
            return max(iterkeys(repn))
    else:
        if repn.linear is not None:
            return 1
        else:
            return 0



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['SOSConstraint']

import sys
import logging
import weakref
from six.moves import zip

from coopr.pyomo.base.component import ActiveComponentData, register_component
from coopr.pyomo.base.sparse_indexed_component import ActiveSparseIndexedComponent
from coopr.pyomo.base.set_types import PositiveIntegers
from coopr.pyomo.base.sets import Set


logger = logging.getLogger('coopr.pyomo')


class _SOSConstraintData(ActiveComponentData): 
    """
    This class defines the data for a single special ordered set.

    Constructor arguments:
        component       The Constraint object that owns this data.

    Public class attributes:
        active          A boolean that is true if this objective is active in the model.
        component       The constraint component.

    Private class attributes:
        _members         SOS member variables.
        _weights         SOS member weights.
        _level           SOS level (Positive Integer)
    """

    __pickle_slots__ = ( '_members', '_weights', '_level')
    __slots__ = __pickle_slots__ + ( '__weakref__', )

    def __init__(self, owner):
        self._level = None
        self._members = []
        self._weights = []
        
        ActiveComponentData.__init__(self, owner)

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        result = super(_SOSConstraintData, self).__getstate__()
        for i in _SOSConstraintData.__pickle_slots__:
            result[i] = getattr(self, i)
        return result

    # Since this class requires no special processing of the state
    # dictionary, it does not need to implement __setstate__()

    def get_weights(self):
        return self._weights

    def get_members(self):
        return self._members

    def get_items(self):
        assert len(self._members) == len(self._weights)
        return list(zip(self._members, self._weights))

    def get_level(self):
        return self._level

    def member_index(self, vardata):
        # it does not suffice to check "if vardata in self._members"
        # because variables hash by .value not by "is"
        member_idx = None
        for idx, member_vardata in enumerate(self._members):
            if vardata is member_vardata:
                member_idx = idx
        return member_idx

    def add_member(self, vardata, weight=None):
        if self.member_index(vardata) is not None:
            raise ValueError("Variable '%s' is already a member of SOSConstraint '%s'" \
                                 % (vardata.cname(True), self.cname(True)))
        self._members.append(vardata)
        if weight is None:
            if len(self._weights) == 0:
                weight = 1
            else:
                weight = max(self._weights)+1
        self._weights.append(weight)

    def remove_member(self, vardata):
        idx = self.member_index(vardata)
        if idx is None:
            raise ValueError("Variable '%s' is not a member of SOSConstraint '%s'" \
                                 % (vardata.cname(True), self.cname(True)))
        del self._members[idx]
        del self._weights[idx]

    def set_weight(self, vardata, weight):
        idx = self.member_index(vardata)
        if idx is None:
            raise ValueError("Variable '%s' is not a member of SOSConstraint '%s'" \
                                 % (vardata.cname(True), self.cname(True)))
        self._weights[idx] = weight
                                 
    def set_level(self, level):
        if level not in PositiveIntegers:
            raise ValueError("SOS Constraint level must be a positive integer")
        self._level = level


class SOSConstraint(ActiveSparseIndexedComponent):
    """
    Represents an SOS-n constraint.

    Usage:
    model.C1 = SOSConstraint(
                             [...],
                             var=VAR,
                             [set=SET OR index=SET],
                             [sos=N OR level=N]
                             [weights=WEIGHTS]
                             )
        [...]   Any number of sets used to index SET
        VAR     The set of variables making up the SOS. Indexed by SET.
        SET     The set used to index VAR. SET is optionally indexed by
                the [...] sets. If SET is not specified, VAR is indexed
                over the set(s) it was defined with.
        N       This constraint is an SOS-N constraint. Defaults to 1.
        WEIGHTS A Param representing the variables weights in the SOS sets.
                A simple counter is used to generate weights when this keyword
                is not used.

    Example:

      model = AbstractModel()
      model.A = Set()
      model.B = Set(A)
      model.X = Set(B)

      model.C1 = SOSConstraint(model.A, var=model.X, set=model.B, sos=1)

    This constraint actually creates one SOS-1 constraint for each
    element of model.A (e.g., if |A| == N, there are N constraints).
    In each constraint, model.X is indexed by the elements of
    model.D[a], where 'a' is the current index of model.A.

      model = AbstractModel()
      model.A = Set()
      model.X = Var(model.A)

      model.C2 = SOSConstraint(var=model.X, sos=2)

    This produces exactly one SOS-2 constraint using all the variables
    in model.X.
    """

    def __new__(cls, *args, **kwds):
        if cls != SOSConstraint:
            return super(SOSConstraint, cls).__new__(cls)
        if args == ():
            return SimpleSOSConstraint.__new__(SimpleSOSConstraint)
        else:
            return IndexedSOSConstraint.__new__(IndexedSOSConstraint)

    def __init__(self, *args, **kwargs):
        # Get the 'var' parameter
        sosVars = kwargs.pop('var', None)
        # Make sure we have a variable
        if sosVars is None:
            raise TypeError("SOSConstraint() requires the 'var' keyword " \
                  "be specified")

        sosWeights = kwargs.pop('weights', None)

        # Get the 'set' or 'index' parameters
        if 'set' in kwargs and 'index' in kwargs:
            raise TypeError("Specify only one of 'set' and 'index' -- " \
                  "they are equivalent parameters")
        sosSet = kwargs.pop('set', None)
        sosSet = kwargs.pop('index', sosSet)

        # Get the 'sos' or 'level' parameters
        if 'sos' in kwargs and 'level' in kwargs:
            raise TypeError("Specify only one of 'sos' and 'level' -- " \
                  "they are equivalent parameters")
        sosLevel = kwargs.pop('sos', None)
        sosLevel = kwargs.pop('level', sosLevel)

        # Make sure sosLevel has been set
        if sosLevel is None:
            raise TypeError("SOSConstraint() requires that either the " \
                  "'sos' or 'level' keyword arguments be set to indicate " \
                  "the type of SOS.")

        # Set member attributes
        self._sosVars = sosVars
        self._sosWeights = sosWeights
        self._sosSet = sosSet
        self._sosLevel = sosLevel

        kwargs.setdefault('ctype', SOSConstraint)
        ActiveSparseIndexedComponent.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        assert data is None # because I don't know why it's an argument

        generate_debug_messages = (__debug__ is True) and \
                                  (logger.isEnabledFor(logging.DEBUG) is True)

        if generate_debug_messages is True:
            logger.debug("Constructing SOSConstraint %s",self.cname(True))

        if self._constructed is True:
            return
        self._constructed = True

        for index in self._index:
            if generate_debug_messages:
                logger.debug("  Constructing "+self.cname(True)+" index "+str(index))
            self.add(index)

    def add(self, index):
        
        if index is None:
            # because SimpleSOSConstraint already makes an _SOSConstraintData instance
            soscondata = self._data[None]
        else:
            soscondata = self._data[index] = _SOSConstraintData(self)
        soscondata.set_level(self._sosLevel)

        if (self._sosSet is None):
            sosSet = self._sosVars.index_set()
        else:
            if index is None:
                sosSet = self._sosSet
            else:
                sosSet = self._sosSet[index]

        weights = None
        if index is None:
            vars = [self._sosVars[idx] for idx in sosSet]
            if self._sosWeights is not None:
                weights = [self._sosWeights[idx] for idx in sosSet]
            else:
                weights = list(i for i,idx in enumerate(sosSet,1))
        else:
            vars = [self._sosVars[idx] for idx in sosSet]
            if self._sosWeights is not None:
                weights = [self._sosWeights[idx] for idx in sosSet]
            else:
                weights = list(i for i,idx in enumerate(sosSet,1))

        for var, weight in zip(vars,weights):
            soscondata.add_member(var,weight)

    # NOTE: the prefix option is ignored
    def pprint(self, ostream=None, verbose=False, prefix=""):
        """TODO"""
        if ostream is None:
            ostream = sys.stdout
        ostream.write("   "+self.cname()+" : ")
        if not self.doc is None:
            ostream.write(self.doc+'\n')
            ostream.write("  ")
        ostream.write("\tSize="+str(len(self._data.keys()))+' ')
        if isinstance(self._index,Set):
            ostream.write("\tIndex= "+self._index.cname(True)+'\n')
        else:
            ostream.write("\n")
        for val in self._data:
            if not val is None:
                ostream.write("\t"+str(val)+'\n')
            ostream.write("\t\tType="+str(self._data[val].get_level())+'\n')
            ostream.write("\t\tMembers= (Weight:Variable)\n")
            for var, weight in zip(self._data[val].get_members(), self._data[val].get_weights()):
                ostream.write("\t\t\t"+str(weight)+':'+var.cname(True)+'\n')


class SimpleSOSConstraint(SOSConstraint, _SOSConstraintData):

    def __init__(self, *args, **kwd):
        _SOSConstraintData.__init__(self, self)
        SOSConstraint.__init__(self, *args, **kwd)
        self._data[None] = self

    # Since this class derives from Component and Component.__getstate__
    # just packs up the entire __dict__ into the state dict, there s
    # nothng special that we need to do here.  We will just defer to the
    # super() get/set state.  Since all of our get/set state methods
    # rely on super() to traverse the MRO, this will automatically pick
    # up both the Component and Data base classes.


class IndexedSOSConstraint(SOSConstraint):

    def __init__(self, *args, **kwds):
        super(IndexedSOSConstraint,self).__init__(*args, **kwds)


register_component(SOSConstraint, "SOS constraint expressions in a model.")




# This Pyomo example is formulated as a python script.
# To run this script execute the following command:
#
# $ coopr_python ipopt_scaling.py
#
# Execution of this script requires that the ipopt
# solver is in the current search path for executables
# on this system. This example was tested using Ipopt
# version 3.10.2

import coopr.environ
from coopr.pyomo import *
from coopr.opt import SolverFactory

### Create the ipopt solver plugin using the ASL interface
solver = 'ipopt'
solver_io = 'nl'
stream_solver = True    # True prints solver output to screen
keepfiles =     False    # True prints intermediate file names (.nl,.sol,...) 
opt = SolverFactory(solver,solver_io=solver_io)

if opt is None:
    print("")
    print("ERROR: Unable to create solver plugin for %s "\
          "using the %s interface" % (solver, solver_io))
    print("")
    exit(1)
###

### Set Ipopt options to accept the scaling_factor suffix
opt.options['nlp_scaling_method'] = 'user-scaling'
###

### Create the example model
model = ConcreteModel()
model.s = Set(initialize=[1,2,3])
model.y = Var(bounds=(1,5),initialize=1.0)
model.x = Var(model.s,bounds=(1,5),initialize=5.0)
model.obj = Objective(expr=model.y*model.x[3]*(model.y+model.x[1]+model.x[2]) + model.x[2])
model.inequality = Constraint(expr=model.y*model.x[1]*model.x[2]*model.x[3] >= 25.0)
model.equality = Constraint(expr=model.y**2 + model.x[1]**2 + model.x[2]**2 + model.x[3]**2 == 40.0)
###

### Declare the scaling_factor suffix 
model.scaling_factor = Suffix(direction=Suffix.EXPORT)
# set objective scaling factor
model.scaling_factor.setValue(model.obj,4.23)
# set variable scaling factor
model.scaling_factor.setValue(model.y,2.0)
model.scaling_factor.setValue(model.x, 5.0)
model.scaling_factor.setValue(model.x[1], 1.5)
# set constraint scaling factor
model.scaling_factor.setValue(model.inequality, 2.0)
model.scaling_factor.setValue(model.equality, 2.0)
###

### Generate the constraint expression trees if necessary
if solver_io != 'nl':
    # only required when not using the ASL interface
    model.preprocess()
###

### Send the model to ipopt and collect the solution
results = opt.solve(model,keepfiles=keepfiles,tee=stream_solver)
# load the results (including any values for previously declared
# IMPORT / IMPORT_EXPORT Suffix components)
model.load(results)
###
